###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == "__main__":

    # First, create a term manager instance.
    tm = TermManager()
    # Create a Bitwuzla options instance.
    options = Options()
    options.set(Option.PRODUCE_MODELS, True)
    # Then, create a Bitwuzla instance.
    bitwuzla = Bitwuzla(tm, options)
    # Create some sorts.
    bv8  = tm.mk_bv_sort(8)
    bv32 = tm.mk_bv_sort(32)
    fp16 = tm.mk_fp_sort(5, 11)
    # Create terms.
    b     = tm.mk_const(tm.mk_bool_sort(), "b")
    bv    = tm.mk_const(bv8, "bv")
    fp    = tm.mk_const(fp16, "fp")
    rm    = tm.mk_const(tm.mk_rm_sort(), "rm")
    fun   = tm.mk_const(tm.mk_fun_sort([bv8, fp16, bv32], fp16), "fun")
    zero  = tm.mk_bv_zero(bv8)
    ones  = tm.mk_bv_ones(tm.mk_bv_sort(23))
    z     = tm.mk_var(bv8, "z")
    q     = tm.mk_var(bv8, "q")
    lambd = tm.mk_term(Kind.LAMBDA, [z, tm.mk_term(Kind.BV_ADD, [z, bv])])
    fpleq = tm.mk_term(
        Kind.FP_LEQ,
        [tm.mk_term(Kind.APPLY,
                  [fun, bv, fp, tm.mk_term(Kind.BV_ZERO_EXTEND, [ones], [9])]),
          fp])
    exists = tm.mk_term(
        Kind.EXISTS,
        [q, tm.mk_term(Kind.EQUAL, [zero, tm.mk_term(Kind.BV_MUL, [bv, q])])])
    # Assert formulas.
    bitwuzla.assert_formula(b)
    bitwuzla.assert_formula(
        tm.mk_term(Kind.EQUAL, [tm.mk_term(Kind.APPLY, [lambd, bv]), zero]))
    bitwuzla.assert_formula(exists)
    bitwuzla.assert_formula(fpleq)

    # Print sort.
    print('Print bit-vector sort of size 32:')
    print('---------------------------------')
    print(f'str(): {bv32}')
    print()

    # Print terms.
    # Note: Hexadecimal bv output format is ignored if the value is not of size
    #       divisible by 4.
    print('Print term:')
    print('-----------')
    print(f'str()      [default]:       {rm}')
    print(f'str()      [bin (ignored)]: {rm.str(2)}')
    print(f'str()      [dec (ignored)]: {rm.str(10)}')
    print(f'str(16)    [hex (ignored)]: {rm.str(16)}')
    print()
    print(f'str()      [default]: {zero}')
    print(f'str()      [bin]:     {zero.str(2)}')
    print(f'str(10)    [dec]:     {zero.str(10)}')
    print(f'str(16)    [hex]:     {zero.str(16)}')
    print()
    print(f'str()      [default]:       {fpleq}')
    print(f'str()      [bin]:           {fpleq.str()}')
    print(f'str(10)    [dec]:           {fpleq.str(10)}')
    print(f'str(16)    [hex (ignored)]: {fpleq.str(16)}')
    print()

    # Print asserted formulas.
    # Note: This uses the default bit-vector output format (binary).
    expected_smt2 = \
            '(set-logic UFBVFP)\n' \
            + '(declare-const b Bool)\n' \
            + '(declare-const bv (_ BitVec 8))\n' \
            + '(declare-const fp (_ FloatingPoint 5 11))\n' \
            + '(declare-fun fun ((_ BitVec 8) (_ FloatingPoint 5 11) ' \
            + '(_ BitVec 32)) (_ FloatingPoint 5 11))\n' \
            + '(assert b)\n' \
            + '(assert (= ((lambda ((z (_ BitVec 8))) (bvadd z bv)) bv) ' \
            + '#b00000000))\n' \
            + '(assert (exists ((q (_ BitVec 8))) (= #b00000000 ' \
            + '(bvmul bv q))))\n' \
            + '(assert (fp.leq (fun bv fp ((_ zero_extend 9) ' \
            + '#b11111111111111111111111)) fp))\n' \
            + '(check-sat)\n' \
            + '(exit)\n'
    res = bitwuzla.print_formula()
    assert res == expected_smt2
    print('Print formula [default (binary) bv output format]:')
    print('--------------------------------------------------')
    print(res)

    # Print asserted formulas using hexadecimal bit-vector output format.
    expected_smt2 = \
            '(set-logic UFBVFP)\n' \
            + '(declare-const b Bool)\n' \
            + '(declare-const bv (_ BitVec 8))\n' \
            + '(declare-const fp (_ FloatingPoint 5 11))\n' \
            + '(declare-fun fun ((_ BitVec 8) (_ FloatingPoint 5 11) ' \
            + '(_ BitVec 32)) (_ FloatingPoint 5 11))\n' \
            + '(assert b)\n' \
            + '(assert (= ((lambda ((z (_ BitVec 8))) (bvadd z bv)) bv) ' \
            + '#x00))\n' \
            + '(assert (exists ((q (_ BitVec 8))) (= #x00 (bvmul bv q))))\n' \
            + '(assert (fp.leq (fun bv fp ((_ zero_extend 9) ' \
            + '#b11111111111111111111111)) fp))\n' \
            + '(check-sat)\n' \
            + '(exit)\n'
    res = bitwuzla.print_formula("smt2", 16)
    assert res == expected_smt2
    print('Print formula [hexadecimal bv output format]:')
    print('--------------------------------------------------')
    print(res)

    # Print asserted formulas using decimal bit-vector output format.
    expected_smt2 = \
            '(set-logic UFBVFP)\n' \
            + '(declare-const b Bool)\n' \
            + '(declare-const bv (_ BitVec 8))\n' \
            + '(declare-const fp (_ FloatingPoint 5 11))\n' \
            + '(declare-fun fun ((_ BitVec 8) (_ FloatingPoint 5 11) ' \
            + '(_ BitVec 32)) (_ FloatingPoint 5 11))\n' \
            + '(assert b)\n' \
            + '(assert (= ((lambda ((z (_ BitVec 8))) (bvadd z bv)) bv) ' \
            + '(_ bv0 8)))\n' \
            + '(assert (exists ((q (_ BitVec 8))) (= (_ bv0 8) ' \
            + '(bvmul bv q))))\n' \
            + '(assert (fp.leq (fun bv fp ((_ zero_extend 9) ' \
            + '(_ bv8388607 23))) fp))\n' \
            + '(check-sat)\n' \
            + '(exit)\n'
    res = bitwuzla.print_formula("smt2", 10)
    assert res == expected_smt2
    print('Print formula [decimal bv output format]:')
    print('---------------------------------------------')
    print(res)

    bitwuzla.check_sat()

    # Print values.
    print('Print value of Boolean predicate:')
    print('---------------------------------')
    fpleq_val = bitwuzla.get_value(fpleq).value()
    print(f'{fpleq}: {fpleq_val} [bool]')
    print()

    print('Print value of bv const:')
    print('------------------------')
    print(f'{bv}: {bitwuzla.get_value(bv).value():>8} [str] (bin)')
    print(f'{bv}: {bitwuzla.get_value(bv).value(10):>8} [str] (dec)')
    print(f'{bv}: {bitwuzla.get_value(bv).value(16):>8} [str] (hex)')
    print()

    print('Print value of RoundingMode const:')
    print('----------------------------------')
    print(f'{rm}: {bitwuzla.get_value(rm).value()} [RoundingMode]')
    print()

    fp_val = bitwuzla.get_value(fp)

    print('Print value of fp const as single bit-vector (base ignored):')
    print('------------------------------------------------------------')
    assert fp_val.value(2, False) == fp_val.value(10, False)
    assert fp_val.value(2, False) == fp_val.value(16, False)
    print(f'{fp}: {fp_val.value(2, False):>16} [str] (bin)')
    print(f'{fp}: {fp_val.value(10, False):>16} [str] (dec [ignored])')
    print(f'{fp}: {fp_val.value(16, False):>16} [str] (hex [ignored])')
    print()

    print('Print value of fp const as list of component bit-vectors:')
    print('---------------------------------------------------------')
    val = fp_val.value(2)
    print(f'{fp}: [{val[0]}, {val[1]:>5}, {val[2]:>11}] [str] (bin)')
    val = fp_val.value(10)
    print(f'{fp}: [{val[0]}, {val[1]:>5}, {val[2]:>11}] [str] (dec)')
    val = fp_val.value(16)
    print(f'{fp}: [{val[0]}, {val[1]:>5}, {val[2]:>11}] [str] (hex)')
    print()



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == '__main__':

    # First, create a Bitwuzla options instance.
    options = Options()

    # Enable model generation, which expects a boolean configuration value.
    options.set(Option.PRODUCE_MODELS, True)

    # Increase the verbosity level, which expects an integer value.
    print(f'Previous verbosity level: {options.get(Option.VERBOSITY)}')
    options.set(Option.VERBOSITY, 2)
    print(f'Current verbosity level: {options.get(Option.VERBOSITY)}')

    # Now configure an option that has modes (a choice of configuration values).
    # First, query which bit-vector solver engine is set.
    print(f'Default bv solver: {options.get(Option.BV_SOLVER)}')
    # Then, select the propagation-based local search engine as solver engine.
    options.set(Option.BV_SOLVER, 'prop')
    print(f'Current engine: {options.get(Option.BV_SOLVER)}')

    # Now, create a Bitwuzla instance.
    tm = TermManager()
    bitwuzla = Bitwuzla(tm, options)
    # Check sat (nothing to solve, input formula is empty).
    result = bitwuzla.check_sat()
    print('Expect: sat')
    print(f'Bitwuzla: {result}')



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == '__main__':

    # First, create a term manager instance.
    tm = TermManager()
    # Create a Bitwuzla options instance.
    options = Options()
    # Then, enable unsat core extraction.
    # Note: Unsat core extraction is disabled by default.
    options.set(Option.PRODUCE_UNSAT_CORES, True)

    # Then, create a Bitwuzla instance.
    bitwuzla = Bitwuzla(tm, options)

    # Create a bit-vector sort of size 2.
    sortbv2 = tm.mk_bv_sort(2)
    # Create a bit-vector sort of size 4.
    sortbv4 = tm.mk_bv_sort(4)
    # Create Float16 floatinf-point sort.
    sortfp16 = tm.mk_fp_sort(5, 11)

    # Create bit-vector variables.
    # (declare-const x0 (_ BitVec 4))
    x0 = tm.mk_const(sortbv4, 'x0')
    # (declare-const x1 (_ BitVec 2))
    x1 = tm.mk_const(sortbv2, 'x1')
    # (declare-const x2 (_ BitVec 2))
    x2 = tm.mk_const(sortbv2, 'x2')
    # (declare-const x3 (_ BitVec 2))
    x3 = tm.mk_const(sortbv2, 'x3')
    # (declare-const x4 Float16)
    x4 = tm.mk_const(sortfp16, 'x4')

    # Create FP positive zero.
    fpzero = tm.mk_fp_pos_zero(sortfp16)
    # Create BV zero of size 4.
    bvzero = tm.mk_bv_zero(sortbv4)

    # (define-fun f0 ((a Float16)) Bool (fp.gt a (_ +zero 5 11)))
    a_f0 = tm.mk_var(sortfp16, 'a_f0')
    f0 = tm.mk_term(Kind.LAMBDA, [a_f0, tm.mk_term(Kind.FP_GT, [a_f0, fpzero])])

    # (define-fun f1 ((a Float16)) (_ BitVec 4) (ite (f0 a) x0 #b0000))
    a_f1 = tm.mk_var(sortfp16, 'a_f1')
    f1   = tm.mk_term(
        Kind.LAMBDA,
        [a_f1,
           tm.mk_term(Kind.ITE, [tm.mk_term(Kind.APPLY, [f0, a_f1]), x0, bvzero])])

    # (define-fun f2 ((a Float16)) (_ BitVec 2) ((_ extract 1 0) (f1 a)))
    a_f2 = tm.mk_var(sortfp16, 'a_f2')
    f2   = tm.mk_term(
        Kind.LAMBDA,
        [a_f2,
           tm.mk_term(Kind.BV_EXTRACT, [tm.mk_term(Kind.APPLY, [f1, a_f2])], [1, 0])])

    # (assert (! (bvult x2 (f2 (_ +zero 5 11))) :named a0))
    a0 = tm.mk_term(Kind.BV_ULT, [x2, tm.mk_term(Kind.APPLY, [f2, fpzero])])
    bitwuzla.assert_formula(a0)

    # (assert (! (= x1 x2 x3) :named a1))
    a1 = tm.mk_term(Kind.EQUAL, [x1, x2, x3])
    bitwuzla.assert_formula(a1)

    # (assert (!(= x4 ((_ to_fp_unsigned 5 11) RNE x3)) :named a2))
    a2 = tm.mk_term(Kind.EQUAL,
                      [x4,
                       tm.mk_term(Kind.FP_TO_FP_FROM_UBV,
                               [tm.mk_rm_value(RoundingMode.RNE), x3],
                               [5, 11])])
    bitwuzla.assert_formula(a2)

    # (check-sat)
    result = bitwuzla.check_sat()
    print('Expect: unsat')
    print(f'Bitwuzla: {result}')

    # (get-unsat-core)
    unsat_core = bitwuzla.get_unsat_core()
    print('Unsat Core:')
    print('{')
    for t in unsat_core:
        print(f' {t}')
    print('}')



#!/usr/bin/env python3
###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import argparse
import os
import subprocess
import shutil
import sys

from pathlib import Path

def info(msg):
    print(f'-- {msg}')

def die(msg):
    sys.exit(f'** configure.py: {msg}')

def patch_mpfr_pc(builddir):
    try:
        if not os.path.exists(builddir):
            os.mkdir(builddir)
        cmd = subprocess.Popen(
                ['pkg-config', '--variable=pcfiledir', 'mpfr'],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = cmd.communicate()
        new_lines = []
        with open(os.path.join(Path(out.decode().strip()), 'mpfr.pc'), 'r') \
                as infile:
            lines = infile.readlines()
            for line in lines:
                if line.startswith('Libs:'):
                    new_lines.append('Requires: gmp >= 5.0\n')
                    line = line.replace('-lgmp', '')
                new_lines.append(line)
        with open(os.path.join(builddir, 'mpfr.pc'), 'w') as outfile:
            outfile.write(''.join(new_lines))
        if 'PKG_CONFIG_PATH' in os.environ:
            os.environ['PKG_CONFIG_PATH'] += os.pathsep + builddir
        else:
            os.environ['PKG_CONFIG_PATH'] = os.pathsep + builddir
    except FileNotFoundError:
        pass

def configure_build(builddir, opts):
    cmd = ['meson']
    if os.path.exists(os.path.join(builddir, 'meson-private', 'build.dat')):
        cmd.append('configure')
        if not opts:
            info(f'{builddir} already up-to-date')
            sys.exit(0)
    else:
        cmd.append('setup')
    cmd.append(builddir)
    cmd.extend(opts)
    info(' '.join(cmd))
    subprocess.run(cmd)
    info(f'compile Bitwuzla with: cd {builddir} && meson compile')

def _feat(val):
    if val is None:
        return 'auto'
    if val:
        return 'enabled'
    return 'disabled'

def _bool(val):
    assert isinstance(val, bool)
    if val:
        return 'true'
    return 'false'

# Can be replaced with argparse.BooleanOptionalAction when Python 3.8 is EOL.
def bool_opt(ap, name, help, dflt=None):
    dest = name.replace('-', '_')
    ap.add_argument(f'--{name}', action='store_true',
                    help=f'enable {help}', default=dflt)
    ap.add_argument(f'--no-{name}', action='store_false', dest=dest,
                    help=f'disable {help}', default=(not dflt))

def main():
    if not os.path.exists('src/main/main.cpp'):
        die('not called from Bitwuzla base directory')

    if shutil.which('meson') is None:
        die('meson not found on system, please install via pip.')

    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    ap.add_argument('buildtype', nargs='?',
                    choices=['release', 'debug', 'debugoptimized'])
    ap.add_argument('-b', '--build-dir', default='build', metavar='DIR',
                    help='build directory')
    ap.add_argument('--prefix',
                    help='install prefix')
    ap.add_argument('--shared', action='store_true',
                    help='shared library')
    ap.add_argument('--static', action='store_true',
                    help='static library')
    bool_opt(ap, 'assertions', 'assertions')
    bool_opt(ap, 'asan', 'address sanitizer')
    bool_opt(ap, 'ubsan', 'undefined behavior sanitizer')
    bool_opt(ap, 'coverage', 'code coverage')
    ap.add_argument('--win64', action='store_true',
                    help='enable cross compilation for 64-bit Windows')
    ap.add_argument('--arm64', action='store_true',
                    help='enable cross compilation for 64-bit ARM')
    bool_opt(ap, 'python', 'python bindings')
    bool_opt(ap, 'testing', 'regression and unit testing')
    bool_opt(ap, 'unit-testing', 'unit testing')
    bool_opt(ap, 'docs', 'documentation')
    ap.add_argument('--wipe', action='store_true',
                    help='delete build directory if it already exists')
    bool_opt(ap, 'cadical', 'CaDiCaL support', True)
    bool_opt(ap, 'kissat', 'Kissat support')
    bool_opt(ap, 'cryptominisat', 'CryptoMiniSat support')
    bool_opt(ap, 'gimsatul', 'Gimsatul support')
    bool_opt(ap,
             'fpexp',
             'support for experimental floating-point formats, ' +
             'all formats except Float16, Float32, Float64 and Float128 are ' +
             'considered experimental (due to known issues in SymFPU), use ' +
             'at your own risk')
    bool_opt(ap, 'aiger', 'AIGER support to print AIGs')
    bool_opt(ap, 'mimalloc', 'Use mimalloc memory allocator')
    args = ap.parse_args()

    build_opts = []
    sanitize = set()
    if args.buildtype:
        build_opts.append(f'-Dbuildtype={args.buildtype}')
    if args.prefix:
        build_opts.append(f'-Dprefix={args.prefix}')
    if args.shared:
        build_opts.append(f'-Ddefault_library=shared')
    if args.static:
        build_opts.append(f'-Ddefault_library=static')
    if args.asan is not None:
        sanitize.add('address' if args.asan else 'none')
    if args.ubsan is not None:
        if args.ubsan:
            sanitize.add('undefined')
        elif 'address' not in sanitize:
            sanitize.add('none')
    if args.assertions is not None:
        build_opts.append(f'-Db_ndebug={_bool(not args.assertions)}')
    if args.testing is not None:
        build_opts.append(f'-Dtesting={_feat(args.testing)}')
    if args.unit_testing is not None:
        build_opts.append(f'-Dunit_testing={_feat(args.unit_testing)}')
    if args.coverage is not None:
        build_opts.append(f'-Db_coverage={_bool(args.coverage)}')
    if args.win64:
        build_opts.append('--cross-file=x86_64-w64-mingw32.txt')
    if args.arm64:
        build_opts.append('--cross-file=x86_64-linux-aarch64.txt')
    if args.python is not None:
        build_opts.append(f'-Dpython={_bool(args.python)}')
    if args.docs is not None:
        build_opts.append(f'-Ddocs={_bool(args.docs)}')
    if sanitize:
        build_opts.append(f'-Db_sanitize={",".join(sanitize)}')
    if args.wipe and os.path.exists(args.build_dir):
        shutil.rmtree(args.build_dir)
    if args.cadical is not None:
        build_opts.append(f'-Dcadical={_bool(args.cadical)}')
    if args.kissat is not None:
        build_opts.append(f'-Dkissat={_bool(args.kissat)}')
    if args.cryptominisat is not None:
        build_opts.append(f'-Dcryptominisat={_bool(args.cryptominisat)}')
    if args.gimsatul is not None:
        build_opts.append(f'-Dgimsatul={_bool(args.gimsatul)}')
    if args.fpexp is not None:
        build_opts.append(f'-Dfpexp={_bool(args.fpexp)}')
    else:
        if args.buildtype and args.buildtype != 'release':
            build_opts.append(f'-Dfpexp=true')
        else:
            build_opts.append(f'-Dfpexp=false')
    if args.aiger is not None:
        build_opts.append(f'-Daiger={_bool(args.aiger)}')
    if args.mimalloc is not None:
        build_opts.append(f'-Dmimalloc={_bool(args.mimalloc)}')

    # workaround until mpfr.pc is fixed from mpfr side
    patch_mpfr_pc(args.build_dir)

    configure_build(args.build_dir, build_opts)

if __name__ == '__main__':
    main()



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import argparse
import os
import shutil

ap = argparse.ArgumentParser()
ap.add_argument('output')
args = ap.parse_args()

if not os.path.isdir(args.output):
    raise RuntimeError(f'Directory `{args.output}` does not exist.')

docs_dir = f'{args.output}-gh'
# Delete directory if it already exists
if os.path.exists(docs_dir):
    print(f'[jekyll-fix] delete {docs_dir}')
    shutil.rmtree(docs_dir)
# Copy directory
print(f'[jekyll-fix] copy {args.output} to {docs_dir}')
src = shutil.copytree(args.output, docs_dir)

# Rename _static to static since Jekyll ignores directory with _ prefix.
src = os.path.join(docs_dir, '_static')
dst = os.path.join(docs_dir, 'static')
if os.path.exists(src):
    if os.path.exists(dst):
        shutil.rmtree(dst)
    print(f'[jekyll-fix] rename {src} to {dst}')
    os.rename(src, dst)
if not os.path.exists(dst):
    raise RuntimeError(f'Directory `{dst}` not found after rename.')

# Rename static/sphinx_javascript_frameworks_compat.js to
# static/sphinx_javascript_frameworks_compat.js
src = os.path.join(dst, '_sphinx_javascript_frameworks_compat.js')
dst = os.path.join(dst, 'sphinx_javascript_frameworks_compat.js')
if os.path.exists(src):
    if os.path.exists(dst):
        shutil.rmtree(dst)
    print(f'[jekyll-fix] rename {src} to {dst}')
    os.rename(src, dst)
if not os.path.exists(dst):
    raise RuntimeError(f'File `{dst}` not found after rename.')


# Delete _sources
src = os.path.join(docs_dir, '_sources')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    shutil.rmtree(src)
# Delete conf.py
src = os.path.join(docs_dir, 'conf.py')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    os.remove(src)
# Delete objects.inv
src = os.path.join(docs_dir, 'objects.inv')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    os.remove(src)
# Delete c/Doxyfile
src = os.path.join(docs_dir, 'c/Doxyfile')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    os.remove(src)
# Delete c/xml
src = os.path.join(docs_dir, 'c/xml')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    shutil.rmtree(src)
# Delete cpp/Doxyfile
src = os.path.join(docs_dir, 'cpp/Doxyfile')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    os.remove(src)
# Delete cpp/xml
src = os.path.join(docs_dir, 'cpp/xml')
if os.path.exists(src):
    print(f'[jekyll-fix] delete {src}')
    shutil.rmtree(src)

# Rename _static to static and _sphinx_javascript_frameworks_compat.js to
# sphinx_javascript_frameworks_compat.js in all documentation files
print(f'[jekyll-fix] rename "_static" and ' +
       '"_sphinx_javascript_frameworks_compat.js" in files')
for root, dirs, files in os.walk(docs_dir):
    for file in files:
        if not file.endswith('.html'):
            continue
        filename = os.path.join(root, file)
        with open(filename, 'r+') as file:
            contents = file.read()
            file.seek(0)
            file.write(contents.replace('_static', 'static').
                replace(
                    '_sphinx_javascript_frameworks_compat.js',
                    'sphinx_javascript_frameworks_compat.js'))
            file.truncate()



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2021 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import os

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import StringList

class TabbedExamples(Directive):
    """Add directive `tabbed-examples` to be used as follows:

        .. tabbed-examples::
            file1
            file2

        The arguments should be relative paths to source files.
        This directive detects the language from the file extension and
        supports the extensions specified in self.exts.
    """

    # Set tab title and language for syntax highlighting
    exts = {
        '.c': {'title': 'C', 'lang': 'c'},
        '.cpp': {'title': 'C++', 'lang': 'c++'},
        '.py': {'title': 'Python', 'lang': 'python'},
        '.smt2': {'title': 'SMT-LIBv2', 'lang': 'smtlib'},
    }

    # The "arguments" are actually the content of the directive
    has_content = True

    def run(self):
        # collect everything in a list of strings
        content = ['.. tabs::', '']

        for file in self.content:
            # detect file extension
            _, ext = os.path.splitext(file)
            if ext in self.exts:
                title = self.exts[ext]['title']
                lang = self.exts[ext]['lang']
            else:
                title = ext
                lang = ext

            # generate tabs
            content.append(f'    .. tab:: {title}')
            content.append(f'')
            content.append(f'        .. literalinclude:: {file}')
            content.append(f'            :language: {lang}')
            content.append(f'            :linenos:')

        # parse the string list
        node = nodes.Element()
        self.state.nested_parse(StringList(content), 0, node)
        return node.children


def setup(app):
    app.setup_extension('sphinx_tabs.tabs')
    app.add_directive("tabbed-examples", TabbedExamples)
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2020 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

"""
The Python API of the SMT solver Bitwuzla.
"""

cimport bitwuzla_api
from libc.stdint cimport uint8_t, int32_t, uint32_t, uint64_t
from libcpp cimport bool as c_bool
from libcpp.optional cimport optional, nullopt, make_optional
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector
from libcpp.memory cimport unique_ptr, shared_ptr
from cpython.ref cimport PyObject
from cython.operator import dereference
import cython

include "enums.pxd"
include "options.pxd"
include "result.pxd"

# -------------------------------------------------------------------------- #

class BitwuzlaException(Exception):
    """Bitwuzla exception class."""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

# --------------------------------------------------------------------------- #
# Utility functions
# --------------------------------------------------------------------------- #

cdef str _to_str(const char *string):
    if string is NULL:
        return None
    cdef bytes py_str = string
    return str(py_str.decode())

cdef Term _term(tm: TermManager, term: bitwuzla_api.Term):
    t = Term()
    t.c_term = term
    t.tm = tm
    return t

cdef bitwuzla_api.Term _cterm(term: Term):
    return term.c_term

cdef list[Term] _terms(tm: TermManager, vector[bitwuzla_api.Term]& c_terms):
    return [_term(tm, t) for t in c_terms]

cdef vector[bitwuzla_api.Term] _term_vec(terms):
    cdef vector[bitwuzla_api.Term] vec
    for t in terms:
        vec.push_back(_cterm(t))
    return vec

cdef Sort _sort(tm: TermManager, sort: bitwuzla_api.Sort):
    s = Sort()
    s.c_sort = sort
    s.tm = tm
    return s

cdef list[Sort] _sorts(tm: TermManager, vector[bitwuzla_api.Sort]& c_sorts):
    return [_sort(tm, s) for s in c_sorts]

cdef bitwuzla_api.Sort _csort(sort: Sort):
    return sort.c_sort

cdef vector[bitwuzla_api.Sort] _sort_vec(list sorts):
    cdef vector[bitwuzla_api.Sort] vec
    for s in sorts:
        vec.push_back(_csort(s))
    return vec

def _check_arg(arg, _type):
    if not isinstance(_type, list):
        _type = [_type]

    if not any([isinstance(arg, _t) for _t in _type]):
        raise ValueError(
                f'Expected type {str(_type)} for argument, but got {type(arg)}')

# -------------------------------------------------------------------------- #
# Library info                                                               #
# -------------------------------------------------------------------------- #

def copyright() -> str:
    """:return: The copyright information.
       :rtype: str
    """
    return bitwuzla_api.copyright().decode()

def version() -> str:
    """:return: The version number.
       :rtype: str
    """
    return bitwuzla_api.version().decode()

def git_id() -> str:
    """:return: The git commit sha.
       :rtype: str
    """
    return bitwuzla_api.git_id().decode()

# --------------------------------------------------------------------------- #
# Sort wrapper
# --------------------------------------------------------------------------- #

cdef class Sort:
    """A Bitwuzla sort."""
    cdef bitwuzla_api.Sort c_sort
    cdef TermManager tm

    def is_null(self) -> bool:
        """Determine if this sort is a null sort.

           :return: True if this sort is a null sort.
        """
        return self.c_sort.is_null()

    def id(self) -> int:
        """Get the id of this sort.

           :return: The sort id.
        """
        return self.c_sort.id()

    def bv_size(self) -> int:
        """Get the size of a bit-vector sort.

           Requires that given sort is a bit-vector sort.

           :return: The size of the bit-vector sort.
        """
        return self.c_sort.bv_size()

    def fp_exp_size(self) -> int:
        """Get the exponent size of a floating-point sort.

           Requires that given sort is a floating-point sort.

           :return: The exponent size of the floating-point sort.
        """
        return self.c_sort.fp_exp_size()

    def fp_sig_size(self) -> int:
        """ Get the significand size of a floating-point sort.

            Requires that given sort is a floating-point sort.

            :return: The significand size of the floating-point sort.
        """
        return self.c_sort.fp_sig_size()

    def array_index(self) -> Sort:
        """Get the index sort of an array sort.

           Requires that given sort is an array sort.

           :return: The index sort of the array sort.
        """
        return _sort(self.tm, self.c_sort.array_index())

    def array_element(self) -> Sort:
        """Get the element sort of an array sort.

           Requires that given sort is an array sort.

           :return: The element sort of the array sort.
        """
        return _sort(self.tm, self.c_sort.array_element())

    def fun_domain(self) -> list[Sort]:
        """Get the domain sorts of a function sort.

           Requires that given sort is a function sort.

           :return: The domain sorts of the function sort.
        """
        return [_sort(self.tm, s) for s in self.c_sort.fun_domain()]

    def fun_codomain(self) -> Sort:
        """Get the codomain sort of a function sort.

           Requires that given sort is a function sort.

           :return: The codomain sort of the function sort.
        """
        return _sort(self.tm, self.c_sort.fun_codomain())

    def fun_arity(self) -> int:
        """Get the arity of a function sort.

           :return: The number of arguments of the function sort.
        """
        return self.c_sort.fun_arity()

    def uninterpreted_symbol(self) -> str:
        """Get the symbol of an uninterpreted sort.

           :return: The symbol.
        """
        symbol = self.c_sort.uninterpreted_symbol()
        return symbol.value().decode() if symbol.has_value() else None

    def is_array(self) -> bool:
        """Determine if this sort is an array sort.

           :return: True if this sort is an array sort.
        """
        return self.c_sort.is_array()

    def is_bool(self) -> bool:
        """Determine if this sort is a Boolean sort.

           :return: True if this sort is a Boolean sort.
        """
        return self.c_sort.is_bool()

    def is_bv(self) -> bool:
        """Determine if this sort is a bit-vector sort.

           :return: True if `sort` is a bit-vector sort.
        """
        return self.c_sort.is_bv()

    def is_fp(self) -> bool:
        """Determine if this sort is a floating-point sort.

           :return: True if this sort is a floating-point sort.
        """
        return self.c_sort.is_fp()

    def is_fun(self) -> bool:
        """Determine if this sort is a function sort.

           :return: True if this sort is a function sort.
        """
        return self.c_sort.is_fun()

    def is_rm(self) -> bool:
        """Determine if this sort is a Roundingmode sort.

           :return: True if this sort is a Roundingmode sort.
        """
        return self.c_sort.is_rm()

    def is_uninterpreted(self) -> bool:
        """Determine if this sort is an uninterpreted sort.

           :return: True if this sort is an uninterpreted sort.
        """
        return self.c_sort.is_uninterpreted()

    def __str__(self) -> str:
        """Get string representation of this sort.

           :return: String representation of this sort.
        """
        return self.c_sort.str().decode()

    def __eq__(self, Sort other) -> bool:
        """Syntactical equality operator.

           :param a: The first sort.
           :param b: The second sort.
           :return: True if the given sorts are equal.
        """
        return self.c_sort == other.c_sort

    def __hash__(self):
        """Hash function for Sort.

           :param sort: The sort.
           :return: The hash value of the sort.
        """
        return self.id()


# --------------------------------------------------------------------------- #
# Term wrapper
# --------------------------------------------------------------------------- #

cdef class Term:
    """A Bitwuzla term."""
    cdef bitwuzla_api.Term c_term
    cdef TermManager tm

    def is_null(self) -> bool:
        """Determine if this term is a null term.

           :return: True if this term is a null term.
        """
        return self.c_term.is_null()

    def id(self) -> int:
        """Get the id of this term.

           :return: The term id.
        """
        return self.c_term.id()

    def kind(self) -> Kind:
        """Get the kind of this term.

           :return: The kind.
        """
        return Kind(self.c_term.kind())

    def sort(self) -> Sort:
        """Get the sort of this term.

           :return: The sort of the term.
        """
        return _sort(self.tm, self.c_term.sort())

    def num_children(self) -> int:
        """Get the number of child terms of this term.

           :return: The number of children of this term.
        """
        return self.c_term.num_children()

    def children(self) -> list[Term]:
        """Get the child terms of this term.

           :return: The children of this term as a vector of terms.
        """
        return _terms(self.tm, self.c_term.children())

    def __getitem__(self, uint64_t index) -> Term:
        """Get child at position `index`.

           :note: Only valid to call if num_children() > 0.

           :param index: The position of the child.
           :return: The child node at position `index`.
        """
        return _term(self.tm, self.c_term[index])

    def num_indices(self) -> int:
        """Get the number of indices of this term.

           :return: The number of indices of this term.
        """
        return self.c_term.num_indices()

    def indices(self) -> list[int]:
        """Get the indices of an indexed term.

           Requires that given term is an indexed term.

           :return: The indices of this term as a vector of indices.
        """
        return [i for i in self.c_term.indices()]

    def symbol(self) -> str:
        """Get the symbol of this term.

           :return: The symbol of this term. `None` if no symbol is defined.
        """
        opt = self.c_term.symbol()
        if opt.has_value():
            return opt.value().get().decode()
        return None

    def is_const(self) -> bool:
        """Determine if this term is a constant.

           :return: True if this term is a constant.
        """
        return self.c_term.is_const()

    def is_variable(self) -> bool:
        """Determine if this term is a variable.

           :return: True if this term is a variable.
        """
        return self.c_term.is_variable()

    def is_value(self) -> bool:
        """Determine if this term is a value.

           :return: True if this term is a value.
        """
        return self.c_term.is_value()

    def is_true(self) -> bool:
        """Determine if this term is Boolean value true.

           :return: True if this term is Boolean value true.
        """
        return self.c_term.is_true()

    def is_false(self) -> bool:
        """Determine if this term is Boolean value false.

           :return: True if this term is Boolean value false.
        """
        return self.c_term.is_false()

    def is_bv_value_zero(self) -> bool:
        """Determine if this term is a bit-vector value representing zero.

           :return: True if this term is a bit-vector zero value.
        """
        return self.c_term.is_bv_value_zero()

    def is_bv_value_one(self) -> bool:
        """Determine if this term is a bit-vector value representing one.

           :return: True if this term is a bit-vector one value.
        """
        return self.c_term.is_bv_value_one()

    def is_bv_value_ones(self) -> bool:
        """Determine if this term is a bit-vector value with all bits set to one.

           :return: True if this term is a bit-vector value with all bits set to one.
        """
        return self.c_term.is_bv_value_ones()

    def is_bv_value_min_signed(self) -> bool:
        """Determine if this term is a bit-vector minimum signed value.

           :return: True if this term is a bit-vector value with the most
                    significant bit set to 1 and all other bits set to 0.
        """
        return self.c_term.is_bv_value_min_signed()

    def is_bv_value_max_signed(self) -> bool:
        """Determine if this term is a bit-vector maximum signed value.

           :return: True if this term is a bit-vector value with the most
                    significant bit set to 0 and all other bits set to 1.
        """
        return self.c_term.is_bv_value_max_signed()

    def is_fp_value_pos_zero(self) -> bool:
        """Determine if this term is a floating-point positive zero (``+zero``)
           value.

           :return: True if this term is a floating-point ``+zero`` value.
        """
        return self.c_term.is_fp_value_pos_zero()

    def is_fp_value_neg_zero(self) -> bool:
        """Determine if this term is a floating-point value negative zero
           (``-zero``).

           :return: True if this term is a floating-point ``-zero``.
        """
        return self.c_term.is_fp_value_neg_zero()

    def is_fp_value_pos_inf(self) -> bool:
        """Determine if this term is a floating-point positive infinity
           (``+oo``) value.

           :return: True if this term is a floating-point ``+oo`` value.
        """
        return self.c_term.is_fp_value_pos_inf()

    def is_fp_value_neg_inf(self) -> bool:
        """Determine if this term is a floating-point negative infinity
           (``-oo``) value.

           :return: True if this term is a floating-point ``-oo`` value.
        """
        return self.c_term.is_fp_value_neg_inf()

    def is_fp_value_nan(self) -> bool:
        """Determine if this term is a floating-point NaN value.

           :return: True if this term is a floating-point NaN value.
        """
        return self.c_term.is_fp_value_nan()

    def is_rm_value_rna(self) -> bool:
        """Determine if this term is a rounding mode ``RNA`` value.

           :return: True if this term is a roundindg mode ``RNA`` value.
        """
        return self.c_term.is_rm_value_rna()

    def is_rm_value_rne(self) -> bool:
        """Determine if this term is a rounding mode ``RNE`` value.

           :return: True if this term is a rounding mode ``RNE`` value.
        """
        return self.c_term.is_rm_value_rne()

    def is_rm_value_rtn(self) -> bool:
        """Determine if this term is a rounding mode ``RTN`` value.

           :return: True if this term is a rounding mode RTN value.
        """
        return self.c_term.is_rm_value_rtn()

    def is_rm_value_rtp(self) -> bool:
        """Determine if this term is a rounding mode ``RTP`` value.

           :return: True if this term is a rounding mode ``RTP`` value.
        """
        return self.c_term.is_rm_value_rtp()

    def is_rm_value_rtz(self) -> bool:
        """Determine if this term is a rounding mode ``RTZ`` value.

           :return: True if this term is a rounding mode ``RTZ`` value.
        """
        return self.c_term.is_rm_value_rtz()

    def value(self, uint8_t base = 2, fp_as_tuple = True):
        """Get value from value term.

           This function is instantiated for types
           - ``bool`` for Boolean values
           - ``RoundingMode`` for rounding mode values
           - ``string`` for bit-vector and floating-point values

           In case of string representations of values this returns the raw
           value string (as opposed to str(), which returns the SMT-LIB v2
           representation of a term). For example, this function returns
           ``"010"`` for a bit-vector value 2 of size 3, while str() returns
           ``"#b010"``.

           :note: The returned string for floating-point values is always the
                  binary IEEE-754 representation of the value (parameter `base`
                  is ignored). Parameter `base` always configures the numeric
                  base for bit-vector values, and for floating-point values in
                  case of the tuple of strings instantiation. It is always
                  ignored for Boolean and RoundingMode values.

           :param base: The numeric base for bit-vector values; ``2`` for
                         binary, ``10`` for decimal, and ``16`` for hexadecimal.
           :param fp_as_tuple: True if a floating-point value is to be
                                represented as a tuple of raw string
                                representations of the sign, exponent and
                                significand bit-vectors. False to represent as
                                the binary IEEE-754 string representation of a
                                single bit-vector.
        """
        sort = self.sort()
        if sort.is_bool():
            return self.c_term.value[c_bool]()
        if sort.is_rm():
            return RoundingMode(self.c_term.value[bitwuzla_api.RoundingMode]())
        if sort.is_fp() and fp_as_tuple:
            return [s.decode()
                    for s in bitwuzla_api.get_fp_value_ieee(self.c_term, base)]
        return self.c_term.value[string](base).decode()

    def str(self, uint8_t base = 2) -> str:
        """Get the SMT-LIB v2 string representation of this term.

           :param base: The base of the string representation of bit-vector
                         values; ``2`` for binary, ``10`` for decimal, and
                         ``16`` for hexadecimal. Always ignored for Boolean and
                         RoundingMode values.
           :return: A string representation of this term.
        """
        return self.c_term.str(base).decode()

    def fp_value_to_real_str(self) -> str:
        """Get the SMT-LIB v2 string representation as a Real of this
           floating-point

           :note: The floating-point values for NaN and positive and negative
                  infinity don't have a representation as a Real and are thus
                  printed as, e.g., ``(fp.to_real (_ NaN 5 11))`` (for Float16).
           :return: A string representation of this term.
        """
        return self.c_term.fp_value_to_real_str().decode()

    def __str__(self) -> str:
        """Get the SMT-LIB v2 string representation of this term.

           :note: This uses the default (binary) output number format for
                  bit-vector values.
           :return: A string representation of this term.
        """
        return self.c_term.str(2).decode()

    def __eq__(self, other: Term) -> bool:
        """Syntactical equality operator.

           :param a: The first term.
           :param b: The second term.
           :return: True if the given terms are equal.
        """
        return self.c_term == other.c_term

    def __hash__(self):
        return self.id()


# --------------------------------------------------------------------------- #
# Options wrapper
# --------------------------------------------------------------------------- #

cdef class Options:
    """A Bitwuzla options configuration."""
    cdef bitwuzla_api.Options c_options

    def is_valid(self, name: str):
        """Determine if given string is a valid short or long option name.
           :param name: The name.
           :return: True if given string is a option name.
        """
        return self.c_options.is_valid(<const string&> name.encode())

    def shrt(self, option: Option) -> str:
        """Get the short name of this option.

           :return: The short name of this option.
        """
        return self.c_options.shrt(option.value).decode()

    def lng(self, option: Option) -> str:
        """Get the long name of this option.

           :return: The long name of this option.
        """
        return self.c_options.lng(option.value).decode()

    def description(self, option: Option) -> str:
        """Get the description of this option.

           :return: The description of this option.
        """
        return self.c_options.description(option.value).decode()

    def modes(self, option: Option) -> list[str]:
        """Get the modes of this option.

           :return: The modes of this option.
        """
        return [m.decode() for m in self.c_options.modes(option.value)]

    def set(self, option, value):
        """ Set option.

            :param option: The option or the long or short name of an option.
            :param value:  The option value.
        """
        cdef bitwuzla_api.Option opt
        if isinstance(option, Option):
            opt = option.value
        elif isinstance(option, str):
            opt = self.option(option).value
        else:
            raise ValueError(f'Invalid option {option}.')

        if isinstance(value, str):
            self.c_options.set(opt, <const string&> value.encode())
        elif isinstance(value, bool) or isinstance(value, int):
            self.c_options.set(opt, <uint64_t?> value)
        else:
            raise ValueError(f'Invalid value type for option {option.value}.')

    def set_args(self, *args):
        """ Set option via command-line.

            Supports the following command line option format:

            Short option names: ::

              -short      ... ["-short"]
              -short=val  ... ["-short=val"]
              -short val  ... ["-short", "val"]

            Long option names: ::

              --long      ... ["--long"]
              --long=val  ... ["--long=val"]
              --long val  ... ["--long", "val"]

            :param args: List of command line options.
        """
        cdef vector[string] opts
        for a in args:
            opts.push_back(str(a).encode())
        self.c_options.set(opts)

    def option(self, name: str) -> Option:
        """ Get the option associated to the given short or long option name.

            :return: The option associated to the given short or long option
                     name.
        """
        return Option(self.c_options.option(name.encode()))

    def get(self, option: Option):
        """ Get the current value of the given option.

            :param option: The option.
            :return: The current value of the given option.
        """
        if self.c_options.is_mode(option.value):
            return self.c_options.get_mode(option.value).decode()
        elif self.c_options.is_numeric(option.value):
            return self.c_options.get(option.value)
        elif self.c_options.is_bool(option.value):
            return self.c_options.get(option.value) != 0

    def is_bool(self, option: Option):
        """ Determine if given option is a Boolean option.

            :param option: The option to query.
            :return: True if `option` is a Boolean option.
        """
        return self.c_options.is_bool(option.value)

    def is_numeric(self, option: Option):
        """ Determine if given option is a numeric option.

            :param option: The option to query.
            :return: True if `option` is a numeric option.
        """
        return self.c_options.is_numeric(option.value)


    def is_mode(self, option: Option):
        """ Determine if given option is an option with a mode.

            :param option: The option to query.
            :return: True if `option` is an option with a mode.
        """
        return self.c_options.is_mode(option.value)

# --------------------------------------------------------------------------- #
# OptionInfo wrapper
# --------------------------------------------------------------------------- #

class OptionInfoKind(Enum):
    """The kind of an OptionInfo.

    """
    BOOL = bitwuzla_api.CppOptionInfoKind.BOOL
    NUMERIC = bitwuzla_api.CppOptionInfoKind.NUMERIC
    MODE = bitwuzla_api.CppOptionInfoKind.MODE
    STRING = bitwuzla_api.CppOptionInfoKind.STRING

cdef class OptionInfo:
    """ The class holding all information about an option.
    """
    cdef unique_ptr[bitwuzla_api.OptionInfo] c_info

    def __init__(self, options: Options, option: Option):
        self.c_info.reset(
                new bitwuzla_api.OptionInfo(options.c_options, option.value))

    def shrt(self):
        """Get the short name of an option.
           :return: The short name, if any, else None.
        """
        return _to_str(self.c_info.get().shrt)

    def lng(self):
        """Get the long name of an option.
           :return: The long name of an option.
        """
        return _to_str(self.c_info.get().lng)

    def kind(self) -> OptionInfoKind:
        """Get the kind of this OptionInfo.
           :return: The kind.
        """
        return OptionInfoKind(self.c_info.get().kind)

    def dflt(self):
        """Get the default value of an option.
           :return: The default value.
        """
        if self.kind() == OptionInfoKind.BOOL:
            return self.c_info.get().value[bitwuzla_api.OptionInfoBool]().dflt
        if self.kind() == OptionInfoKind.MODE:
            return self.c_info.get().value[
                    bitwuzla_api.OptionInfoMode]().dflt.decode()
        if self.kind() == OptionInfoKind.STRING:
            return self.c_info.get().value[
                    bitwuzla_api.OptionInfoString]().dflt.decode()
        assert self.kind() == OptionInfoKind.NUMERIC
        return self.c_info.get().value[bitwuzla_api.OptionInfoNumeric]().dflt

    def cur(self):
        """Get the current value of an option.
           :return: The current default value.
        """
        if self.kind() == OptionInfoKind.BOOL:
            return self.c_info.get().value[bitwuzla_api.OptionInfoBool]().cur
        if self.kind() == OptionInfoKind.MODE:
            return self.c_info.get().value[
                    bitwuzla_api.OptionInfoMode]().cur.decode()
        if self.kind() == OptionInfoKind.STRING:
            return self.c_info.get().value[
                    bitwuzla_api.OptionInfoString]().cur.decode()
        assert self.kind() == OptionInfoKind.NUMERIC
        return self.c_info.get().value[bitwuzla_api.OptionInfoNumeric]().cur

    def min(self):
        """Get the minimum value of a numeric option.
           :return: The minimum value.
        """
        if self.kind() == OptionInfoKind.NUMERIC:
            return self.c_info.get().value[bitwuzla_api.OptionInfoNumeric]().min
        return None

    def max(self):
        """Get the maximum value of a numeric option.
           :return: The maximum value.
        """
        if self.kind() == OptionInfoKind.NUMERIC:
            return self.c_info.get().value[bitwuzla_api.OptionInfoNumeric]().max
        return None

    def modes(self):
        """Get the available modes of an option with modes.
           :return: The modes.
        """
        if self.kind() == OptionInfoKind.MODE:
            return [
                    _to_str(m) for m in
                    self.c_info.get().value[bitwuzla_api.OptionInfoMode]().modes
                   ]
        return None

    def description(self):
        """Get the description of an option.
           :return: The description.
        """
        return _to_str(self.c_info.get().description)

# --------------------------------------------------------------------------- #
# TermManager wrapper
# --------------------------------------------------------------------------- #

cdef class TermManager:
    """The term manager is responsible for the creation and management of sorts
       and terms. Sorts and terms are tied to a specific term manager instance
       and can be shared between multiple solver instances that have the same
       term manager.
    """
    cdef shared_ptr[bitwuzla_api.TermManager] c_tm
    cdef bitwuzla_api.TermManager* c_tm_ptr

    def __init__(self):
        self.c_tm.reset(new bitwuzla_api.TermManager())
        self.c_tm_ptr = self.c_tm.get()

    def __eq__(self, other: TermManager):
        return self.c_tm_ptr == other.c_tm_ptr

    # ----------------------------------------------------------------------- #
    # Sort functions
    # ----------------------------------------------------------------------- #

    def mk_bool_sort(self) -> Sort:
        """Create a Boolean sort.

           :return: Sort of type Boolean.
        """
        return _sort(self, self.c_tm_ptr.mk_bool_sort())

    def mk_bv_sort(self, size: uint64_t) -> Sort:
        """Create bit-vector sort of size ``size``.

           :param size: Bit width.
           :return:  Bit-vector sort of size ``size``.
        """
        return _sort(self, self.c_tm_ptr.mk_bv_sort(size))

    def mk_array_sort(self, index: Sort, elem: Sort) -> Sort:
        """Create array sort with given index and element sorts.

           :param index: The sort of the array index.
           :param elem: The sort of the array elements.
           :return:  Array sort.
          """
        return _sort(self, self.c_tm_ptr.mk_array_sort(index.c_sort, elem.c_sort))

    def mk_fun_sort(self, domain: list[Sort], codomain: Sort) -> Sort:
        """Create function sort with given domain and codomain.

           :param domain: A list of all the function arguments' sorts.
           :param codomain: The sort of the function's return value.
           :return:  Function sort, which maps ``domain`` to ``codomain``.
          """
        return _sort(self, self.c_tm_ptr.mk_fun_sort(_sort_vec(domain),
                                               _csort(codomain)))

    def mk_fp_sort(self, exp_size: uint64_t, sig_size: uint64_t) -> Sort:
        """Create a floating-point sort with given exponent size ``exp_size``
           and significand size ``sig_size``.

           :param exp_size: Exponent size.
           :param sig_size: Significand size.
           :return: Floating-point sort.
        """
        return _sort(self, self.c_tm_ptr.mk_fp_sort(exp_size, sig_size))

    def mk_rm_sort(self, ) -> Sort:
        """Create a rounding mode sort.

           :return: Rounding mode sort.
        """
        return _sort(self, self.c_tm_ptr.mk_rm_sort())

    def mk_uninterpreted_sort(self, symbol: str = None) -> Sort:
        """Create an uninterpreted sort.

           :param symbol: The symbol of the sort.
           :return: Uninterpreted Sort.

           .. note::
                Only 0-arity uninterpreted sorts are supported.
        """
        cdef optional[string] opt = nullopt
        if symbol:
            opt = <string?> symbol.encode()
        return _sort(self,
                self.c_tm_ptr.mk_uninterpreted_sort(
                    <optional[const string]?> opt))


    # ----------------------------------------------------------------------- #
    # Value functions
    # ----------------------------------------------------------------------- #

    def mk_true(self, ) -> Term:
        """mk_true()

        Create true value.

        :return: A term representing true.
        :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_true())

    def mk_false(self, ) -> Term:
        """mk_false()

        Create false value.

        :return: A term representing false.
        :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_false())

    def mk_bv_value(self, sort: Sort, value, *args) -> Term:
        """mk_bv_value(sort: Sort, value: int) -> Term
           mk_bv_value(sort: Sort, value: str, base: int) -> Term

           Create bit-vector representing ``value`` of given ``sort``.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort
           :param value: An integer representing the value.
           :type value: int

           Create bit-vector representing ``value`` of given ``sort`` and
           ``base``.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort
           :param value: A string representing the value.
           :type value: str
           :param base: The numerical base of the string representation (``2``
                        for binary, ``10`` for decimal, ``16`` for hexadecimal).
           :type base: int

           :return: A term representing the bit-vector value.
           :rtype: BitwuzlaTerm
        """
        if isinstance(value, str):
            if not args:
                raise ValueError('expected base')
            return _term(self,
                    self.c_tm_ptr.mk_bv_value(
                        sort.c_sort, value.encode(), <uint8_t> int(args[0])))
        if args:
            raise ValueError('unexpected base')
        return _term(self,
                self.c_tm_ptr.mk_bv_value(sort.c_sort,
                                            str(value).encode(),
                                            10))

    def mk_bv_zero(self, sort: Sort) -> Term:
        """mk_bv_zero(sort)

           Create a bit-vector value zero with ``sort``.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort

           :return: A term representing the bit-vector value zero of given sort.
           :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_bv_zero(sort.c_sort))

    def mk_bv_one(self, sort: Sort) -> Term:
        """mk_bv_one(sort)

           Create a bit-vector value one with ``sort``.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort

           :return: A term representing the bit-vector value one of given sort.
           :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_bv_one(sort.c_sort))

    def mk_bv_ones(self, sort: Sort) -> Term:
        """mk_bv_ones(sort)

           Create a bit-vector value with ``sort`` where all bits are set to 1.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort

           :return: A term representing the bit-vector value of given sort
                    where all bits are set to 1.
           :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_bv_ones(sort.c_sort))

    def mk_bv_min_signed(self, sort: Sort) -> Term:
        """mk_bv_min_signed(sort)

           Create a bit-vector minimum signed value.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort

           :return: A term representing the bit-vector value of given sort
                    where the MSB is set to 1 and all remaining bits are set to
                    0.
           :rtype: BitwuzlaTerm

        """
        return _term(self, self.c_tm_ptr.mk_bv_min_signed(sort.c_sort))

    def mk_bv_max_signed(self, sort: Sort) -> Term:
        """mk_bv_max_signed(sort)

           Create a bit-vector maximum signed value.

           :param sort: Bit-vector sort.
           :type sort: BitwuzlaSort

           :return: A term representing the bit-vector value of given sort
                    where the MSB is set to 0 and all remaining bits are set to
                    1.
           :rtype: BitwuzlaTerm
        """
        return _term(self, self.c_tm_ptr.mk_bv_max_signed(sort.c_sort))

    def mk_fp_value(self, *args) -> Term:
        """mk_fp_value(sign: Term, exponent: Term, significand: Term) -> Term
           mk_fp_value(sort: Sort, rm: Term, real: str) -> Term
           mk_fp_value(sort: Sort, rm: Term, num: str, den: str) -> Term

           Create a floating-point value from its IEEE 754 standard
           representation given as three bit-vector values representing the
           sign bit, the exponent and the significand.

           :param sign: Bit-vector value term representing the sign bit.
           :param exponent: Bit-vector value term representing the exponent.
           :param significand: Bit-vector value term representing the
                               significand.

           Create a floating-point value from its real representation, given as
           a decimal string, with respect to given rounding mode.

           :param sort: Floating-point sort.
           :param rm: Rounding mode term.
           :param real: The decimal string representing a real value

           Create a floating-point value from its rational representation,
           given as a two decimal strings representing the numerator and
           denominator, with respect to given rounding mode.

           :param sort: Floating-point sort.
           :param rm: Rounding mode term.
           :param num: The decimal string representing the numerator.
           :param den: The decimal string representing the denominator.

           :return: A term representing the floating-point value.

           .. seealso::
             :func:`~bitwuzla.TermManager.mk_bv_value` for the supported value
             format for ``sign``, ``exponent``, and ``significand``.
        """
        if isinstance(args[0], Sort):
            _check_arg(args[1], Term)
            if len(args) == 4:
                _check_arg(args[2], [str, int])
                _check_arg(args[3], [str, int])
                return _term(self, self.c_tm_ptr.mk_fp_value(
                                _csort(args[0]),
                                _cterm(args[1]),
                                str(args[2]).encode(),
                                str(args[3]).encode()))
            elif len(args) != 3:
                raise ValueError('Invalid number of arguments')
            _check_arg(args[2], [str, int, float])
            return _term(self, self.c_tm_ptr.mk_fp_value(
                            _csort(args[0]),
                            _cterm(args[1]),
                            <const string&> str(args[2]).encode()))
        elif isinstance(args[0], Term):
            _check_arg(args[1], Term)
            _check_arg(args[2], Term)
            return _term(self, self.c_tm_ptr.mk_fp_value(
                            _cterm(args[0]), _cterm(args[1]), _cterm(args[2])))
        else:
            raise ValueError('Invalid arguments')

    def mk_fp_pos_zero(self, sort: Sort) -> Term:
        """Create a floating-point positive zero value (SMT-LIB: `+zero`).

           :param sort: Floating-point sort.
           :return: A term representing the floating-point positive zero value
                    of given floating-point sort.
        """
        return _term(self, self.c_tm_ptr.mk_fp_pos_zero(_csort(sort)))

    def mk_fp_neg_zero(self, sort: Sort) -> Term:
        """Create a floating-point negative zero value (SMT-LIB: `-zero`).

           :param sort: Floating-point sort.
           :return: A term representing the floating-point negative zero value
                    of given floating-point sort.
        """
        return _term(self, self.c_tm_ptr.mk_fp_neg_zero(_csort(sort)))

    def mk_fp_pos_inf(self, sort: Sort) -> Term:
        """Create a floating-point positive infinity value (SMT-LIB: `+oo`).

           :param sort: Floating-point sort.
           :return: A term representing the floating-point positive infinity
                    value of given floating-point sort.
        """
        return _term(self, self.c_tm_ptr.mk_fp_pos_inf(_csort(sort)))

    def mk_fp_neg_inf(self, sort: Sort) -> Term:
        """Create a floating-point negative infinity value (SMT-LIB: `-oo`).

           :param sort: Floating-point sort.
           :return: A term representing the floating-point negative infinity
                    value of given floating-point sort.
        """
        return _term(self, self.c_tm_ptr.mk_fp_neg_inf(_csort(sort)))

    def mk_fp_nan(self, sort: Sort) -> Term:
        """Create a floating-point NaN value.

           :param sort: Floating-point sort.
           :return: A term representing the floating-point NaN value of given
                    floating-point sort.
        """
        return _term(self, self.c_tm_ptr.mk_fp_nan(_csort(sort)))

    def mk_rm_value(self, rm: RoundingMode) -> Term:
        """Create a rounding mode value.

           :param rm: Rounding mode.
           :return: A term representing the rounding mode value.
        """
        return _term(self, self.c_tm_ptr.mk_rm_value(rm.value))


    # ----------------------------------------------------------------------- #
    # Term functions
    # ----------------------------------------------------------------------- #

    def mk_const(self, sort: Sort, symbol: str = None) -> Term:
        """Create a (first-order) constant of given sort with given symbol.

           :param sort: The sort of the constant.
           :param symbol: The symbol of the constant.
           :return: A term of kind :class:`~bitwuzla.Kind.CONSTANT`,
                    representing the constant.
        """
        cdef optional[string] opt = nullopt
        if symbol:
            opt = <string?> symbol.encode()
        return _term(self, self.c_tm_ptr.mk_const(
                        _csort(sort), <optional[const string]?> opt))

    def mk_const_array(self, sort: Sort, term: Term) -> Term:
        """Create a one-dimensional constant array of given sort, initialized
           with given value.

           :param sort: The sort of the array.
           :param term: The term to initialize the elements of the array with.
           :return: A term of kind :class:`~bitwuzla.Kind.CONST_ARRAY`,
                    representing a constant array of given sort.
        """
        return _term(self,
                     self.c_tm_ptr.mk_const_array(_csort(sort), _cterm(term)))

    def mk_var(self, sort: Sort, symbol: str = None) -> Term:
        """Create a (first-order) variable of given sort with given symbol.

           :param sort: The sort of the variable.
           :param symbol: The symbol of the variable.
           :return: A term of kind :class:`~bitwuzla.Kind.VARIABLE`,
                    representing the variable.
        """
        cdef optional[string] opt = nullopt
        if symbol:
            opt = <string?> symbol.encode()
        return _term(self, self.c_tm_ptr.mk_var(
                        _csort(sort), <optional[const string]?> opt))

    def mk_term(self, kind: Kind,
                terms: list[Term],
                indices: list[int] = []) -> Term:
        """Create a term of given kind with the given argument terms.

           :param kind: The operator kind.
           :param terms: The argument terms.
           :param indices: The indices of this term, empty if not indexed.
           :return: A term representing an operation of given kind.
        """
        return _term(self, self.c_tm_ptr.mk_term(kind.value,
                                             _term_vec(terms),
                                             indices))

    # ----------------------------------------------------------------------- #
    # Term substitution
    # ----------------------------------------------------------------------- #

    def substitute_term(self, term: Term, substs: dict[Term, Term]) -> Term:
        """Substitute a set terms in a given term. The substitutions to perform
           are represented as map from keys to be substituted with their
           corresponding values in the given term.

           :param term: The term in which the terms are to be substituted.
           :param substs: The substitution map.
           :return: The resulting term from this substitution.
        """
        cdef unordered_map[bitwuzla_api.Term, bitwuzla_api.Term] c_substs
        for k,v in substs.items(): c_substs[_cterm(k)] = _cterm(v)
        return _term(self, self.c_tm_ptr.substitute_term(term.c_term, c_substs))

    def substitute_terms(self,
                         terms: list[Term],
                         substs: dict[Term, Term]) -> list[Term]:
        """Substitute a set of terms in a set of given terms. The substitutions
           to perform are represented as map from keys to be substituted with
           their corresponding values in the given terms.

           The terms in `terms` are replaced with the terms resulting from these
           substitutions.

           :param terms: The terms in which the terms are to be substituted.
           :param substs: The substitution map.
        """
        cdef unordered_map[bitwuzla_api.Term, bitwuzla_api.Term] c_substs
        cdef vector[bitwuzla_api.Term] c_terms = _term_vec(terms)
        for k,v in substs.items(): c_substs[_cterm(k)] = _cterm(v)
        self.c_tm_ptr.substitute_terms(c_terms, c_substs)
        return _terms(self, c_terms)


# --------------------------------------------------------------------------- #
# Bitwuzla wrapper
# --------------------------------------------------------------------------- #

cdef class Bitwuzla:
    """A Bitwuzla solver instance."""
    cdef shared_ptr[bitwuzla_api.Bitwuzla] c_bitwuzla
    cdef unique_ptr[bitwuzla_api.PyTerminator] c_terminator
    cdef TermManager tm

    @staticmethod
    @cython.cfunc
    def from_shared_ptr(tm: TermManager(), _sptr: shared_ptr[bitwuzla_api.Bitwuzla]) -> Bitwuzla:
        """Wrap shared_ptr[bitwuzla_api.Bitwuzla] into Bitwuzla instance.
        """
        b: Bitwuzla = Bitwuzla.__new__(Bitwuzla)
        b.c_bitwuzla = _sptr
        b.tm = tm
        return b

    def __init__(self, tm: TermManager, options: Options = Options()):
        """Constructor.

           :param options: The options configuration of the solver instance.
        """
        self.c_bitwuzla.reset(
            new bitwuzla_api.Bitwuzla(dereference(tm.c_tm.get()),
                                                  options.c_options))
        self.tm = tm

    def configure_terminator(self, callback: callable):
        """Set a termination callback.

           Use this function to force Bitwuzla to prematurely terminate if
           callback returns True. The callback object needs to be a callable,
           i.e., it is a function or class that implements the builtin method
           __call__().

           For example: ::

             import time
             class TimeLimitTerminator:
                def __init__(self, time_limit):
                    self.start_time = time.time()
                    self.time_limit = time_limit

                def __call__(self):
                    # Terminate after self.time_limit seconds passed
                    return time.time() - self.start_time > self.time_limit

             bitwuzla = Bitwuzla(tm, options)
             bitwuzla.configure_terminator(lambda: True)            # immediately terminate
             bitwuzla.configure_terminator(TimeLimitTerminator(1))  # terminate after 1s
             bitwuzla.configure_terminator(TimeLimitTerminator(10)) # terminate after 10s

           :param callback: A callable Python object.
        """
        self.c_terminator.reset(
            new bitwuzla_api.PyTerminator(<PyObject*> callback))
        self.c_bitwuzla.get().configure_terminator(
            <bitwuzla_api.Terminator*> self.c_terminator.get())

    def push(self, nlevels: uint32_t = 1):
        """Push new assertion levels.

           :param levels: Number of assertion levels to create.
           :type levels: int
        """
        self.c_bitwuzla.get().push(nlevels)

    def pop(self, nlevels: uint32_t = 1):
        """Pop assertion levels.

           :param levels: Number of assertion levels to pop.
           :type levels: int
        """
        self.c_bitwuzla.get().pop(nlevels)

    def assert_formula(self, *formulas: Term):
        """Assert one or more formulas.

           :param formulas: One or more Boolean terms.
        """
        for f in formulas:
            self.c_bitwuzla.get().assert_formula(_cterm(f))


    def get_assertions(self) -> list[Term]:
        """Get currently asserted formulas.
           :return: List of current assertions.
        """
        return _terms(self.tm, self.c_bitwuzla.get().get_assertions())

    def check_sat(self, *assumptions: Term) -> Result:
        """Check satisfiability of asserted formulas under possibly given
           assumptions.

           :param assumptions: Zero or more Boolean terms.

           :return: - :class:`~bitwuzla.Result.SAT` if the input formula is
                      satisfiable (under possibly given assumptions)
                    - :class:`~bitwuzla.Result.UNSAT` if it is unsatisfiable
                    - :class:`~bitwuzla.Result.UNKNOWN` otherwise

           .. seealso::
               :func:`~bitwuzla.Bitwuzla.assert_formula`,
               :func:`~bitwuzla.Bitwuzla.get_value`,
               :func:`~bitwuzla.Bitwuzla.get_unsat_core`,
               :func:`~bitwuzla.Bitwuzla.get_unsat_assumptions`,
        """
        return Result(self.c_bitwuzla.get().check_sat(_term_vec(assumptions)))

    def is_unsat_assumption(self, Term term) -> bool:
        """Determine if given assumption is unsat.

           Unsat assumptions are those assumptions that force an
           input formula to become unsatisfiable.

           See :func:`~bitwuzla.Bitwuzla.check_sat`.

           :param term: Boolean term.
           :return:  True if assumption is unsat, False otherwise.
        """
        return self.c_bitwuzla.get().is_unsat_assumption(_cterm(term))

    def get_unsat_assumptions(self) -> list[Term]:
        """Return list of unsatisfiable assumptions previously added via
           :func:`~bitwuzla.Bitwuzla.check_sat`.

           Requires that the last :func:`~bitwuzla.Bitwuzla.check_sat` call
           returned `~bitwuzla.Result.UNSAT`.

           :return:  List of unsatisfiable assumptions
        """
        return _terms(self.tm, self.c_bitwuzla.get().get_unsat_assumptions())

    def get_unsat_core(self) -> list[Term]:
        """Return list of unsatisfiable assertions previously added via
           :func:`~bitwuzla.Bitwuzla.assert_formula`.

           Requires that the last :func:`~bitwuzla.Bitwuzla.check_sat` call
           returned :class:`~bitwuzla.Result.UNSAT`.

           :return:  list of unsatisfiable assertions
        """
        return _terms(self.tm, self.c_bitwuzla.get().get_unsat_core())

    def simplify(self):
        """Simplify current set of input assertions.

           .. note::
               Each call to :func:`~bitwuzla.Bitwuzla.check_sat`
               simplifies the input formula as a preprocessing step.
               It is not necessary to call this function explicitly in the
               general case.
        """
        self.c_bitwuzla.get().simplify()

    def simplify_term(self, term: Term):
        """Simplify the given term.

           .. note::
               Each call to :func:`~bitwuzla.Bitwuzla.check_sat`
               simplifies the input formula as a preprocessing step.
        """
        return _term(self.tm, self.c_bitwuzla.get().simplify(_cterm(term)))

    def get_value(self, term: Term) -> Term:
        """Get model value of term.

           Requires that the last :func:`~bitwuzla.Bitwuzla.check_sat` call
           returned `~bitwuzla.Result.SAT`.

           :return: Term representing the model value of `term`.
        """
        return _term(self.tm, self.c_bitwuzla.get().get_value(_cterm(term)))

    def get_interpolant(self, A: list[Term]) -> Term:
        """ Get a term representing the interpolant I given the current set of
            assertions, partitioned into partitions A and B such that
            `(and A B)` is unsat and `(=> A I)` and `(=> I (not B))`.

            Partition A is the given set of assertions, partition B consists of
            the remaining assertions that are not in A.

            Requires that the last :func:`~bitwuzla.Bitwuzla.check_sat()` query
            returned `~bitwuzla.Result.UNSAT`.

            .. note:: Assertions in A and B must be currently asserted formulas.

            :param A: The set of formulas representing partition A. This must be
                      a strict subset of the set of current assertions.

            :return: Interpolant I such that `(=> A I)` and `(=> I (not B))`.

            .. warning:: The signature of this function is experimental and may
                         change in future versions.
        """
        return _term(
                self.tm, self.c_bitwuzla.get().get_interpolant(_term_vec(A)))

    def get_interpolants(self, A: list[list[Term]]) -> list[Term]:
        """ Get an inductive sequence of interpolants
            :math:`\\langle I_1, \\ldots, I_n \\rangle`
            given the current set of assertions :math:`F`
            and a sequence of partitions.

            The sequence of partitions is given as a list of set increments
            :math:`\\{A_1, \\ldots, A_n\\}` of asserted formulas
            :math:`F = \{F_1, F_2, \ldots, F_n\}`, which expands into sets of
            partitions

            .. math::

                 \\{(A_1, B_1), (A_2, B_2), \\ldots, (A_n, B_n)\\}

            such that

            .. math::

                 A_1 &= F_1 \\\\
                 A_2 &= F_1\,\cup\,F_2 \\\\
                 \ldots & \\\\
                 A_n & = F_1\\,\\cup\,F_2\\,\\cup\\,\\ldots\\,\\cup\\,F_n \\\\

            and

            .. math::

                 B_i = F \\setminus A_i \\text{ with } A_i \\wedge B_i \\models \\bot.

            The resulting sequence of interpolants is inductive, i.e., it holds
            that :math:`(I_i \\wedge F_{i+1}) \\rightarrow I_{i+1}`.

            Requires that the last :func:`~bitwuzla.Bitwuzla.check_sat()` query
            returned `~bitwuzla.Result.UNSAT`.

            .. note:: Assertions in :math:`A_i` must be currently asserted
                      formulas.

            .. note:: Current SAT state must be unsat.

            :param A: The set of A increments.

            :return: The interpolation sequence.
            .. warning:: The signature of this function is experimental and may
                         change in future versions.
        """
        cdef vector[vector[bitwuzla_api.Term]] vec
        for v in A: vec.push_back(_term_vec(v))
        return _terms(self.tm, self.c_bitwuzla.get().get_interpolants(vec))

    def print_formula(self, fmt: str = 'smt2', uint8_t base = 2) -> str:
        """Get the current input formula as a string.

           :param fmt: The output format for printing the formula. Currently,
                        only `"smt2"` for the SMT-LIB v2 format is supported.
           :param base: The base of the string representation of bit-vector
                        values; ``2`` for binary, ``10`` for decimal, and
                        ``16`` for hexadecimal. Always ignored for Boolean and
                        RoundingMode values.
           :return: The current input formula as a string in the given format.
        """
        cdef bitwuzla_api.stringstream c_ss
        cdef unique_ptr[bitwuzla_api.set_bv_format] c_bv_fmt

        c_bv_fmt.reset(new bitwuzla_api.set_bv_format(base))

        c_ss << dereference(c_bv_fmt.get())
        self.c_bitwuzla.get().print_formula(c_ss, <const string&> fmt.encode())
        return c_ss.to_string().decode()

    def print_unsat_core(self, fmt: str = 'smt2', uint8_t base = 2) -> str:
        """Get the current unsat core as benchmark as a string.

           :param fmt: The output format for printing the formula. Currently,
                        only `"smt2"` for the SMT-LIB v2 format is supported.
           :param base: The base of the string representation of bit-vector
                        values; ``2`` for binary, ``10`` for decimal, and
                        ``16`` for hexadecimal. Always ignored for Boolean and
                        RoundingMode values.
           :return: The current input formula as a string in the given format.
        """
        cdef bitwuzla_api.stringstream c_ss
        cdef unique_ptr[bitwuzla_api.set_bv_format] c_bv_fmt

        c_bv_fmt.reset(new bitwuzla_api.set_bv_format(base))

        c_ss << dereference(c_bv_fmt.get())
        self.c_bitwuzla.get().print_unsat_core(c_ss,
                                               <const string&> fmt.encode())
        return c_ss.to_string().decode()

    def statistics(self) -> dict[str, str]:
        """Get current statistics.

           :return: A map of strings of statistics entries, maps statistic name
                    to value.
        """
        return {_to_str(k): _to_str(v)
                for [k, v] in self.c_bitwuzla.get().statistics()}

    def term_mgr(self) -> TermManager:
        tm = TermManager()
        tm.c_tm_ptr = &self.c_bitwuzla.get().term_mgr()
        return tm

# --------------------------------------------------------------------------- #
# Parser
# --------------------------------------------------------------------------- #

cdef class Parser:
    """The Bitwuzla parser."""
    cdef unique_ptr[bitwuzla_api.Parser] c_parser
    cdef Options options
    cdef TermManager tm
    cdef unique_ptr[bitwuzla_api.PyTerminator] c_terminator

    def __init__(self,
                 tm: TermManager,
                 options: Options,
                 language = "smt2",
                 uint8_t base = 2):
        """Constructor.

           :note: The parser creates and owns the associated Bitwuzla instance.
           :param options:  The configuration options for the Bitwuzla instance
                            (created by the parser).
           :param language: The format of the input.
           :param base:     The base of the string representation of bit-vector
                            values; ``2`` for binary, ``10`` for decimal, and
                            ``16`` for hexadecimal. Always ignored for Boolean
                            and RoundingMode values.
        """
        cdef unique_ptr[bitwuzla_api.set_bv_format] c_bv_fmt
        self.options = options
        self.tm = tm
        c_bv_fmt.reset(new bitwuzla_api.set_bv_format(base))
        bitwuzla_api.cout << dereference(c_bv_fmt.get())
        self.c_parser.reset(
                new bitwuzla_api.Parser(
                    dereference(tm.c_tm.get()),
                    options.c_options,
                    <const string&> str(language).encode(),
                    &bitwuzla_api.cout))

    def configure_terminator(self, callback: callable):
        """Set a termination callback.

           Use this function to force Bitwuzla to prematurely terminate if
           callback returns True. The callback object needs to be a callable,
           i.e., it is a function or class that implements the builtin method
           __call__().

           For example: ::

             import time
             class TimeLimitTerminator:
                def __init__(self, time_limit):
                    self.start_time = time.time()
                    self.time_limit = time_limit

                def __call__(self):
                    # Terminate after self.time_limit seconds passed
                    return time.time() - self.start_time > self.time_limit

             parser = Parser(tm, options)
             parser.configure_terminator(lambda: True)            # immediately terminate
             parser.configure_terminator(TimeLimitTerminator(1))  # terminate after 1s
             parser.configure_terminator(TimeLimitTerminator(10)) # terminate after 10s

           :param callback: A callable Python object.
        """
        self.c_terminator.reset(
            new bitwuzla_api.PyTerminator(<PyObject*> callback))
        self.c_parser.get().configure_terminator(
            <bitwuzla_api.Terminator*> self.c_terminator.get())

    def configure_auto_print_model(self, value: bool):
        """Enable or disable the automatic printing of the model after each
           satisfiable query.

           Enable to automatically print the model after every sat query.
           Must be enabled to automatically print models for BTOR2 input (does
           not provide a command to print the model like `(get-model)` in
           SMT-LIB2). False (default) configures the standard behavior for
           SMT-LIB2 input (print model only after a `(get-model)` command).

           :note: By default, auto printing of the model is disabled.
           :param value: True to enable auto printing of the model.
        """
        self.c_parser.get().configure_auto_print_model(value)

    def parse(self,
              iinput,
              parse_only: bool = False,
              parse_file: bool = True) -> bool:
        """Parse input, either from a file or from a string.

           :param input:      The name of the input file if ``parse_file`` is
                              ``True``, else a string with the input.
           :param parse_only: ``True`` to only parse without issuing calls to
                              check_sat.
           :param parse_file: ``True`` to parse an input file with the given
                              name ``iinput``, ``False`` to parse from
                              ``iinput`` as a string input.
           :raise BitwuzlaException: On parse error.
           :note: Parameter `parse_only` is redundant for BTOR2 input, its the
                  only available mode for BTOR2 (due to the language not
                  supporting "commands" as in SMT2).
        """
        self.c_parser.get().parse(
                <const string&> str(iinput).encode(), parse_only, parse_file)

    def parse_term(self, iinput) -> Term:
        """Parse term from string.

           :param input: The input string.
           :return:      The parsed term.
           :raise BitwuzlaException: On parse error.
        """
        return _term(self.tm, self.c_parser.get().parse_term(
                <const string&> str(iinput).encode()))

    def parse_sort(self, iinput) -> Sort:
        """Parse sort from string.

           :param input: The input string.
           :return:      The parsed sort.
           :raise BitwuzlaException: On parse error.
        """
        return _sort(self.tm, self.c_parser.get().parse_sort(
                <const string&> str(iinput).encode()))

    def get_declared_sorts(self) -> list[Sort]:
        """Get the current set of (user-)declared sort symbols.

           :note: Corresponds to the sorts declared via SMT-LIB command
                  ``declare-sort``. Will always return an empty list for BTOR2
                  input.
           :return: The declared sorts.
        """
        return _sorts(self.tm, self.c_parser.get().get_declared_sorts())

    def get_declared_funs(self) -> list[Term]:
        """Get the current set of (user-)declared function symbols.

           :note: Corresponds to the function symbols declared via SMT-LIB
                  commands ``declare-const`` and ``declare-fun``.
           :return: The declared function symbols.
        """
        return _terms(self.tm, self.c_parser.get().get_declared_funs())

    def bitwuzla(self) -> Bitwuzla:
        """Get the associated Bitwuzla instance.

           :return: The Bitwuzla instance.
        """
        return Bitwuzla.from_shared_ptr(self.tm, self.c_parser.get().bitwuzla())

    def statistics(self) -> dict[str, str]:
        """Get current statistics.

           :return: A map of strings of statistics entries, maps statistic name
                    to value.
        """
        return {_to_str(k): _to_str(v)
                for [k, v] in self.c_parser.get().statistics()}



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == '__main__':

    # First, create a term manager instance.
    tm = TermManager()
    # Create a Bitwuzla options instance.
    options = Options()
    # (set-option :produce-models true)
    options.set(Option.PRODUCE_MODELS, False)

    # Then, create a Bitwuzla instance.
    bitwuzla = Bitwuzla(tm, options)

    # Create a bit-vector sort of size 3.
    sortbv3 = tm.mk_bv_sort(3)

    # (declare-const x (_ BitVec 3))
    x = tm.mk_const(sortbv3, 'x')

    # (assert (= x #b010))
    bitwuzla.assert_formula(
        tm.mk_term(Kind.EQUAL, [x, tm.mk_bv_value(sortbv3, 2)]))
    # (check-sat)
    result = bitwuzla.check_sat()
    print('Expect: sat')
    print(f'Bitwuzla: {result}')

    # (set-option :produce-models true)
    options.set(Option.PRODUCE_MODELS, True)

    # (reset)
    # Note: Bitwuzla does not provide an explicit API function for reset since
    #       this is achieved by simply discarding the current Bitwuzla instance
    #       and creating a new one.
    bitwuzla = Bitwuzla(tm, options)

    # (declare-const a ( Array (_ BitVec 2) (_ BitVec 3)))
    sortbv2 = tm.mk_bv_sort(2)
    a       = tm.mk_const(tm.mk_array_sort(sortbv2, sortbv3), 'a')

    # (assert (= x #b011))
    bitwuzla.assert_formula(
        tm.mk_term(Kind.EQUAL, [x, tm.mk_bv_value(sortbv3, 3)]))
    # (assert (= x (select a #b01)))
    bitwuzla.assert_formula(tm.mk_term(
        Kind.EQUAL,
        [x, tm.mk_term(Kind.ARRAY_SELECT, [a, tm.mk_bv_value(sortbv2, 1)])]))
    # (check-sat)
    result = bitwuzla.check_sat()
    print('Expect: sat')
    print(f'Bitwuzla: {result}')
    # (get-model)
    print('(')
    print(f'  (define-fun {x.symbol()} () {x.sort()} {bitwuzla.get_value(x)} )')
    print(f'  (define-fun {a.symbol()} () {a.sort()} {bitwuzla.get_value(a)} )')
    print(')')



#
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == "__main__":

    # First, create a term manager instance.
    tm = TermManager()
    # Create a Bitwuzla options instance.
    options = Options()
    # Then, enable model generation.
    options.set(Option.PRODUCE_MODELS, True)
    # Now, for illustration purposes, we enable CaDiCaL as SAT solver
    # (CaDiCaL is already configured by default).
    # Note: This will silently fall back to one of the compiled in SAT solvers
    #       if the selected solver is not compiled in.
    options.set(Option.SAT_SOLVER, 'cadical')
    # Then, create a Bitwuzla instance.
    bitwuzla = Bitwuzla(tm, options)

    # Create bit-vector sorts of size 4 and 8.
    sortbv4 = tm.mk_bv_sort(4)
    sortbv8 = tm.mk_bv_sort(8)
    # Create function sort.
    sortfun = tm.mk_fun_sort([sortbv8, sortbv4], sortbv8)
    # Create array sort.
    sortarr = tm.mk_array_sort(sortbv8, sortbv8)

    # Create two bit-vector constants of that sort.
    x = tm.mk_const(sortbv8, "x")
    y = tm.mk_const(sortbv8, "y")
    # Create fun const.
    f = tm.mk_const(sortfun, "f")
    # Create array const.
    a = tm.mk_const(sortarr, "a")
    # Create bit-vector values one and two of the same sort.
    one = tm.mk_bv_one(sortbv8)
    # Alternatively, you can create bit-vector value one with:
    # one = tm.mk_bv_value(sortbv8, "1", 2)
    # one = tm.mk_bv_value(sortbv8, 1)
    two = tm.mk_bv_value(sortbv8, 2)

    # (bvsdiv x (_ bv2 8))
    sdiv = tm.mk_term(Kind.BV_SDIV, [x, two])
    # (bvashr y (_ bv1 8))
    ashr = tm.mk_term(Kind.BV_ASHR, [y, one])
    # ((_ extract 3 0) (bvsdiv x (_ bv2 8)))
    sdive = tm.mk_term(Kind.BV_EXTRACT, [sdiv], [3, 0])
    # ((_ extract 3 0) (bvashr x (_ bv1 8)))
    ashre = tm.mk_term(Kind.BV_EXTRACT, [ashr], [3, 0])

    # (assert
    #     (distinct
    #         ((_ extract 3 0) (bvsdiv x (_ bv2 8)))
    #         ((_ extract 3 0) (bvashr y (_ bv1 8)))))
    bitwuzla.assert_formula(tm.mk_term(Kind.DISTINCT, [sdive, ashre]))
    # (assert (= (f x ((_ extract 6 3) x)) y))
    bitwuzla.assert_formula(tm.mk_term(
        Kind.EQUAL,
        [tm.mk_term(Kind.APPLY, [f, x, tm.mk_term(Kind.BV_EXTRACT, [x], [6, 3])]),
         y]))
    # (assert (= (select a x) y))
    bitwuzla.assert_formula(
        tm.mk_term(Kind.EQUAL, [tm.mk_term(Kind.ARRAY_SELECT, [a, x]), y]))

    # (check-sat)
    result = bitwuzla.check_sat()
    print('Expect: sat')
    print(f'Bitwuzla: {result}')

    # Print model in SMT-LIBv2 format.
    print('Model:\n(')
    for term in [x, y, f, a]:
        sort = term.sort()
        symbol = term.symbol()
        print(f'  (define-fun {symbol if symbol else "@t" + term.id()} (',
              end = '')
        if sort.is_fun():
            value = bitwuzla.get_value(term)
            assert value.kind() == Kind.LAMBDA
            assert value.num_children() == 2
            while value[1].kind() == Kind.LAMBDA:
                assert value[0].is_variable()
                print(f'({value[0]} {value[0].sort()}) ',
                      end = '')
                value = value[1]
            assert value[0].is_variable()
            print(f'({value[0]} {value[0].sort()})) ' \
                  + f'{sort.fun_codomain()} {value[1]})')
        else:
            print(f') {sort} {bitwuzla.get_value(term)})')
    print(')')
    print()

    # Print value for x, y, f and a.
    # Both x and y are bit-vector terms and their value is a bit-vector
    # value that can be printed via Term.value().
    print(f'value of x: {bitwuzla.get_value(x).value(2)}')
    print(f'value of y: {bitwuzla.get_value(y).value(2)}')
    print()
    # f and a, on the other hand, are a function and array term, respectively.
    # The value of these terms is not a value term: for f, it is a lambda term,
    # and the value of a is represented as a store term. Thus we cannot use
    # Term.value(), but we can print the value of the terms via Term.str().
    print('str() representation of value of f:')
    print(bitwuzla.get_value(f))
    print()
    print('str() representation of value of a:')
    print(bitwuzla.get_value(a))
    print()
    # Note that the assignment string of bit-vector terms is given as the
    # pure assignment string, either in binary, hexadecimal or decimal format,
    # whereas Term.str() prints the value in SMT-LIB2 format (in the given
    # bit-vector output number format, binary by default).
    print(f'str() representation of value of x: {bitwuzla.get_value(x)}')
    print(f'str() representation of value of y: {bitwuzla.get_value(y)}')
    print()

    # Query value of bit-vector term that does not occur in the input formula
    v = bitwuzla.get_value(tm.mk_term(Kind.BV_MUL, [x, x]))
    print(f'value of v = x * x: {v.value(2)}')



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import os

from sphinx.directives.code import LiteralInclude

class IncludeFile(LiteralInclude):

    def run(self):
        assert self.arguments

        rel_path= os.path.relpath(self.env.config.lig_build_dir,
                                  start=self.env.config.lig_source_dir)

        filename = os.path.join(rel_path, self.arguments[0])
        if filename:
            self.arguments[0] = filename
        return super().run()

def setup(app):
    app.add_config_value('lig_source_dir', '', 'env')
    app.add_config_value('lig_build_dir', '', 'env')
    app.add_directive('literalincludegen', IncludeFile)
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True
    }



#!/usr/bin/env python3
###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import argparse
import re
import subprocess

def get_version():
    major = 0
    minor = 0
    patch = 0
    prerelease = None
    with open('meson.build', 'r') as infile:
        content = infile.read()
        m = re.search(r'  version:\s\'(.*)\',', content)
        if m is None:
            raise ValueError('version string not found in `meson.build`')

        try:
            version, prerelease = m.group(1).split('-', 1)
        except ValueError:
            version, prerelease = m.group(1), None

        version = version.split('.')

        if len(version) != 3:
            raise ValueError('invalid format for version string, ' +
                             'expected `format major.minor.patch[-prerelease]`')
        major, minor, patch = version

    return int(major),int(minor),int(patch),prerelease

def update_version(major, minor, patch, prerelease):
    version = f'{major}.{minor}.{patch}'
    if prerelease:
        version += '-' + prerelease
    with open('meson.build', 'r') as infile:
        content = infile.read()
        content = re.sub(r'  version:\s\'(.*)\'', f'  version: \'{version}\'', content)
    with open('meson.build', 'w') as outfile:
        outfile.write(content)
    if not prerelease:
        with open('NEWS.md', 'r') as infile:
            content = infile.read()
            pattern = re.compile(r'## News for version [0-9]\.[0-9]+\.[0-9]+\n')
            m = pattern.search(content)
            versionstr = f'## News for version {version}\n'
            if m and m.group(0) != versionstr:
                print(f'[warning] did not find expected version string: {versionstr}')
                pattern = r'^This file collects a summary.*\n'
                m = re.search(pattern, content, flags = re.MULTILINE)
                mstr = m.group(0) if m else ''
                content = re.sub(
                        pattern,
                        f'{mstr}\n{versionstr}',
                        content,
                        flags = re.MULTILINE)
        with open('NEWS.md', 'w') as outfile:
             outfile.write(content)


def create_commit(msg):
    subprocess.check_call(['git', 'commit', 'meson.build', 'NEWS.md', '-m', msg])

def create_git_tag(tag):
    subprocess.check_call(['git', 'tag', tag])

def get_commit_id():
    proc = subprocess.run(['git', 'rev-parse', 'HEAD'],
                          capture_output=True, check=True)
    return proc.stdout.decode().strip()

def print_diff(commit_id):
    subprocess.check_call(['git', 'show', f'{commit_id}..HEAD'])

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('bump', choices=['major', 'minor', 'patch'])
    args = ap.parse_args()

    major, minor, patch, prerelease = get_version()
    commit_id = get_commit_id()

    # Bump specified version
    if args.bump == 'major':
        major, minor, patch = major + 1, 0, 0
    elif args.bump == 'minor':
        minor, patch = minor + 1, 0
    else:
        assert args.bump == 'patch'
        patch = patch + 1

    update_version(major, minor, patch, None)
    create_commit(f'Bump version to {major}.{minor}.{patch}.')
    tag = f'{major}.{minor}.{patch}'
    create_git_tag(tag)
    update_version(major, minor, patch, 'dev')
    create_commit(f'Start post-release version {major}.{minor}.{patch}-{prerelease}.')

    print_diff(commit_id)
    print(f'Push commits and tag with: `git push && git push origin {tag}`')

if __name__ == '__main__':
    main()



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == "__main__":

    # No options configured, create Bitwuzla instance with default options.
    tm = TermManager()
    bitwuzla = Bitwuzla(tm)

    # Create a bit-vector sort of size 1.
    sortbv1 = tm.mk_bv_sort(1)
    # Create a bit-vector sort of size 2
    sortbv2 = tm.mk_bv_sort(2)

    # Create bit-vector variables.
    # (declare-const o0 (_ BitVec 1))
    o0 = tm.mk_const(sortbv1, 'o0')
    # (declare-const o1 (_ BitVec 1))
    o1 = tm.mk_const(sortbv1, 'o1')
    # (declare-const s0 (_ BitVec 2))
    s0 = tm.mk_const(sortbv2, 's0')
    # (declare-const s1 (_ BitVec 2))
    s1 = tm.mk_const(sortbv2, 's1')
    # (declare-const s2 (_ BitVec 2))
    s2 = tm.mk_const(sortbv2, 's2')
    # (declare-const goal (_ BitVec 2))
    goal = tm.mk_const(sortbv2, 'goal')

    # Create bit-vector values zero, one, three.
    zero  = tm.mk_bv_zero(sortbv2)
    one1  = tm.mk_bv_one(sortbv1)
    one2  = tm.mk_bv_one(sortbv2)
    three = tm.mk_bv_value(sortbv2, 3)

    # Add some assertions.
    bitwuzla.assert_formula(tm.mk_term(Kind.EQUAL, [s0, zero]))
    bitwuzla.assert_formula(tm.mk_term(Kind.EQUAL, [goal, three]))

    # (check-sat-assuming ((= s0 goal)))
    result = bitwuzla.check_sat(tm.mk_term(Kind.EQUAL, [s0, goal]))
    print('Expect: unsat')
    print(f'Bitwuzla: {result}')

    # (assert (= s1 (ite (= o0 (_ sortbv1 1)) (bvadd s0 one) s0)))
    bitwuzla.assert_formula(tm.mk_term(Kind.EQUAL,
                                    [s1,
                                     tm.mk_term(Kind.ITE,
                                             [tm.mk_term(Kind.EQUAL, [o0, one1]),
                                              tm.mk_term(Kind.BV_ADD, [s0, one2]),
                                              s0])]))

    # (check-sat-assuming ((= s1 goal)))
    result = bitwuzla.check_sat(tm.mk_term(Kind.EQUAL, [s1, goal]))
    print('Expect: unsat')
    print(f'Bitwuzla: {result}')

    # (assert (= s2 (ite (= o1 (_ sortbv1 1)) (bvadd s1 one) s1)))
    bitwuzla.assert_formula(tm.mk_term(Kind.EQUAL,
                                     [s2,
                                      tm.mk_term(Kind.ITE,
                                              [tm.mk_term(Kind.EQUAL, [o1, one1]),
                                               tm.mk_term(Kind.BV_ADD, [s1, one2]),
                                               s1])]))

    # (check-sat-assuming ((= s2 goal)))
    result = bitwuzla.check_sat(tm.mk_term(Kind.EQUAL, [s2, goal]))
    print('Expect: unsat')
    print(f'Bitwuzla: {result}')



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2020 by the authors listed in the AUTHORS file at
# https://github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https://github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import BitwuzlaException
from cpython.ref cimport PyObject
from libc.stdint cimport uint8_t, uint32_t, uint64_t, int64_t
from libcpp cimport bool
from libcpp.map cimport map
from libcpp.unordered_map cimport unordered_map
from libcpp.memory cimport shared_ptr
from libcpp.optional cimport optional
from libcpp.string cimport string
from libcpp.vector cimport vector

cdef extern from "<functional>" namespace "std" nogil:
    cdef cppclass reference_wrapper[T]:
        T& get() const

cdef extern from "<iostream>" namespace "std":
    ostream cout
    cdef cppclass ostream:
        ostream &operator << (set_bv_format)

cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream(ostream):
        string to_string "str" () const
        stringstream &operator << (Result)

# Extract C++ exception message.
cdef extern from *:
    """
    #include <bitwuzla/cpp/bitwuzla.h>

    const char*
    get_err_msg()
    {
      try
      {
        throw;
      }
      catch (const bitwuzla::Exception& e)
      {
        return e.what();
      }
    }
    """
    cdef const char* get_err_msg()

cdef inline int raise_error() except *:
    msg = get_err_msg().decode().split("', ")
    raise BitwuzlaException(msg[1] if len(msg) > 1 else msg[0])

# Extract FP value as tuple of strings
cdef extern from *:
    """
    #include <bitwuzla/cpp/bitwuzla.h>

    std::vector<std::string>
    get_fp_value_ieee(const bitwuzla::Term& term, uint8_t base)
    {
      auto val =
        term.value<std::tuple<std::string, std::string, std::string>>(base);
      return {std::get<0>(val), std::get<1>(val), std::get<2>(val)};
    }
    """
    cdef vector[string] get_fp_value_ieee(const Term& term, uint8_t base)

# Terminator wrapper
cdef extern from "terminator.h":
    cdef cppclass PyTerminator:
        PyTerminator(PyObject* terminator)
        bool terminate()


# Bitwuzla C++ API
cdef extern from "bitwuzla/cpp/bitwuzla.h" namespace "bitwuzla":

    string copyright() except +raise_error
    string version() except +raise_error
    string git_id() except +raise_error

    cdef enum class Kind:
        pass

    cdef enum class Result:
        pass

    cdef enum class RoundingMode:
        pass

    cdef enum class Option:
        pass

    cdef cppclass set_bv_format:
        set_bv_format(uint8_t base) except +raise_error
        pass

    cdef cppclass Options:
        Options() except +raise_error
        bool is_valid(const string& name) except +raise_error
        bool is_bool(Option option) except +raise_error
        bool is_numeric(Option option) except +raise_error
        bool is_mode(Option option) except +raise_error
        const char *shrt(Option option) except +raise_error
        const char *lng(Option option) except +raise_error
        const char *description(Option option) except +raise_error
        vector[string] modes(Option option) except +raise_error
        Option option(const char *name) except +raise_error
        void set(Option option, uint64_t value) except +raise_error
        void set(Option option, const string &mode) except +raise_error
        #void set(const string &lng, const string &value) except +raise_error
        void set(const vector[string] &args) except +raise_error
        uint64_t get(Option option) except +raise_error
        const string &get_mode(Option option) except +raise_error

    cdef enum class CppOptionInfoKind "bitwuzla::OptionInfo::Kind":
        BOOL,
        NUMERIC,
        MODE,
        STRING

    cdef cppclass OptionInfoBool "bitwuzla::OptionInfo::Bool":
        bool cur
        bool dflt

    cdef cppclass OptionInfoNumeric "bitwuzla::OptionInfo::Numeric":
        uint64_t cur
        uint64_t dflt
        uint64_t min
        uint64_t max

    cdef cppclass OptionInfoMode "bitwuzla::OptionInfo::Mode":
        string cur
        string dflt
        vector[string] modes

    cdef cppclass OptionInfoString "bitwuzla::OptionInfo::String":
        string cur
        string dflt

    cdef cppclass OptionInfo:
        OptionInfo()
        OptionInfo(Options options, Option option) except +raise_error
        CppOptionInfoKind kind
        const char* lng
        const char* shrt
        const char* description
        OptionInfoBool value[OptionInfoBool]() except +raise_error
        OptionInfoNumeric value[OptionInfoNumeric]() except +raise_error
        OptionInfoMode value[OptionInfoMode]() except +raise_error
        OptionInfoString value[OptionInfoString]() except +raise_error

    cdef cppclass Sort:
        Sort() except +raise_error
        bool is_null() except +raise_error
        uint64_t id() except +raise_error
        uint64_t bv_size() except +raise_error
        uint64_t fp_exp_size() except +raise_error
        uint64_t fp_sig_size() except +raise_error
        Sort array_index() except +raise_error
        Sort array_element() except +raise_error
        vector[Sort] fun_domain() except +raise_error
        Sort fun_codomain() except +raise_error
        size_t fun_arity() except +raise_error
        optional[string] uninterpreted_symbol() except +raise_error
        bool is_array() except +raise_error
        bool is_bool() except +raise_error
        bool is_bv() except +raise_error
        bool is_fp() except +raise_error
        bool is_fun() except +raise_error
        bool is_rm() except +raise_error
        bool is_uninterpreted() except +raise_error
        string str() except +raise_error
        bool operator==(const Sort&) except +raise_error

    cdef cppclass Term:
        Term() except +raise_error
        bool is_null() except +raise_error
        uint64_t id() except +raise_error
        Kind kind() except +raise_error
        Sort sort() except +raise_error
        size_t num_children() except +raise_error
        vector[Term] children() except +raise_error
        Term operator[](size_t index) except +raise_error
        size_t num_indices() except +raise_error
        vector[uint64_t] indices() except +raise_error
        optional[reference_wrapper[const string]] symbol() except +raise_error
        bool is_const() except +raise_error
        bool is_variable() except +raise_error
        bool is_value() except +raise_error
        bool is_true() except +raise_error
        bool is_false() except +raise_error
        bool is_bv_value_zero() except +raise_error
        bool is_bv_value_one() except +raise_error
        bool is_bv_value_ones() except +raise_error
        bool is_bv_value_min_signed() except +raise_error
        bool is_bv_value_max_signed() except +raise_error
        bool is_fp_value_pos_zero() except +raise_error
        bool is_fp_value_neg_zero() except +raise_error
        bool is_fp_value_pos_inf() except +raise_error
        bool is_fp_value_neg_inf() except +raise_error
        bool is_fp_value_nan() except +raise_error
        bool is_rm_value_rna() except +raise_error
        bool is_rm_value_rne() except +raise_error
        bool is_rm_value_rtn() except +raise_error
        bool is_rm_value_rtp() except +raise_error
        bool is_rm_value_rtz() except +raise_error
        string str(uint8_t base) except +raise_error
        string fp_value_to_real_str() except +raise_error
        bool value[bool](uint8_t base) except +raise_error
        RoundingMode value[RoundingMode](uint8_t base) except +raise_error
        string value[string](uint8_t base) except +raise_error
        bool operator==(const Term&) except +raise_error

    cdef cppclass Terminator:
        pass

    cdef cppclass TermManager:
        Sort mk_array_sort(const Sort &index,
                           const Sort &element) except +raise_error
        Sort mk_bool_sort() except +raise_error
        Sort mk_bv_sort(uint64_t size) except +raise_error
        Sort mk_fp_sort(uint64_t exp_size,
                        uint64_t sig_size) except +raise_error
        Sort mk_fun_sort(const vector[Sort] &domain,
                         const Sort &codomain) except +raise_error
        Sort mk_rm_sort() except +raise_error
        Sort mk_uninterpreted_sort(
                optional[const string] symbol) except +raise_error

        Term mk_true() except +raise_error
        Term mk_false() except +raise_error
        Term mk_bv_zero(const Sort &sort) except +raise_error
        Term mk_bv_one(const Sort &sort) except +raise_error
        Term mk_bv_ones(const Sort &sort) except +raise_error
        Term mk_bv_min_signed(const Sort &sort) except +raise_error
        Term mk_bv_max_signed(const Sort &sort) except +raise_error
        Term mk_bv_value(const Sort &sort,
                         const string &value, uint8_t base) except +raise_error
        Term mk_fp_pos_zero(const Sort &sort) except +raise_error
        Term mk_fp_neg_zero(const Sort &sort) except +raise_error
        Term mk_fp_pos_inf(const Sort &sort) except +raise_error
        Term mk_fp_neg_inf(const Sort &sort) except +raise_error
        Term mk_fp_nan(const Sort &sort) except +raise_error
        Term mk_fp_value(const Term &bv_sign,
                         const Term &bv_exponent,
                         const Term &bv_significand) except +raise_error
        Term mk_fp_value(const Sort &sort,
                         const Term &rm,
                         const string &real) except +raise_error
        Term mk_fp_value(const Sort &sort,
                         const Term &rm,
                         const string &num,
                         const string &den) except +raise_error
        Term mk_const_array(const Sort &sort,
                            const Term &term) except +raise_error
        Term mk_rm_value(RoundingMode rm) except +raise_error
        Term mk_term(Kind kind,
                     const vector[Term] &args,
                     const vector[uint64_t] &indices) except +raise_error
        Term mk_const(const Sort &sort,
                      optional[const string] symbol) except +raise_error
        Term mk_var(const Sort &sort,
                    optional[const string] symbol) except +raise_error

        Term substitute_term(const Term &term,
                             const unordered_map[Term, Term] &map) \
                                except +raise_error

        void substitute_terms(vector[Term] &terms,
                              const unordered_map[Term, Term] &map) \
                                except +raise_error

    cdef cppclass Bitwuzla:
        Bitwuzla(TermManager& tm, const Options &options) except +raise_error
        void configure_terminator(Terminator *terminator) except +raise_error
        void push(uint32_t nlevels) except +raise_error
        void pop(uint32_t nlevels) except +raise_error
        void assert_formula(const Term &term) except +raise_error
        vector[Term] get_assertions() except +raise_error;
        bool is_unsat_assumption(const Term &term) except +raise_error
        vector[Term] get_unsat_assumptions() except +raise_error
        vector[Term] get_unsat_core() except +raise_error
        void simplify() except +raise_error
        Term simplify(const Term& term) except +raise_error
        Result check_sat(const vector[Term] &assumptions) except +raise_error
        Term get_value(const Term &term) except +raise_error
        Term get_interpolant(const vector[Term] &A) except +raise_error
        vector[Term] get_interpolants(const vector[vector[Term]] &A) except +raise_error
        void print_formula(ostream& outfile, string& fmt) except +raise_error
        void print_unsat_core(ostream& outfile, string& fmt) except +raise_error
        map[string, string] statistics() except +raise_error
        TermManager& term_mgr() except +raise_error


cdef extern from "bitwuzla/cpp/parser.h" namespace "bitwuzla::parser":

    cdef cppclass Parser:
        Parser(TermManager& tm,
               Options& options,
               const string& language,
               ostream* out) except +raise_error
        void configure_terminator(Terminator *terminator) except +raise_error
        void configure_auto_print_model(bool value) except +raise_error
        void parse(const string& infile_name, bool parse_only, bool parse_file) except +raise_error
        Term parse_term(const string& iinput) except +raise_error
        Sort parse_sort(const string& iinput) except +raise_error
        vector[Sort] get_declared_sorts() except +raise_error
        vector[Term] get_declared_funs() except +raise_error
        string error_msg() except +raise_error
        shared_ptr[Bitwuzla] bitwuzla() except +raise_error
        map[string, string] statistics() except +raise_error

# -------------------------------------------------------------------------- #



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == '__main__':

    # First, create a Bitwuzla options instance.
    options = Options()

    # Set some options to illustrate current vs default value.
    options.set(Option.PRODUCE_MODELS, True)
    options.set(Option.VERBOSITY, 2)
    options.set(Option.BV_SOLVER, 'prop')

    # Then iterate over all available configuration options and extract info.
    for opt in Option:
        info = OptionInfo(options, opt)
        print(f'  long name:   {info.lng()}')
        print(f'  short name:  {info.shrt() if info.shrt() else "-"}')
        print('  kind:        ', end = '')
        if info.kind() == OptionInfoKind.BOOL:
            print('bool')
            print('  values:')
            print(f'  + current:   {info.cur()}')
            print(f'  + default:   {info.dflt()}')
        elif info.kind() == OptionInfoKind.NUMERIC:
            print('numeric')
            print('  values:')
            print(f'  + current:   {info.cur()}')
            print(f'  + default:   {info.dflt()}')
            print(f'  + min:       {info.min()}')
            print(f'  + max:       {info.max()}')
        elif info.kind() == OptionInfoKind.MODE:
            print('modes')
            print('  values:')
            print(f'  + current:   {info.cur()}')
            print(f'  + default:   {info.dflt()}')
            print(f'  + modes:     {info.modes()}')
        else:
            assert info.kind() == OptionInfoKind.STRING
            print('string')
            print('  values:')
            print(f'  + current:   {info.cur()}')
            print(f'  + default:   {info.dflt()}')
        print(f'  description: {info.description()}\n')



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2023 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

import sys

from bitwuzla import *

if __name__ == '__main__':

    # First, create a term manager instance.
    tm = TermManager()
    # Create a Bitwuzla options instance.
    options = Options()

    # We will parse example file `smt2/quickstart.smt2`.
    # Create parser instance.
    parser = Parser(tm, options)

    # Now parse the input file.
    print('Expect: sat')
    print('Bitwuzla: ', end='')
    sys.stdout.flush()
    res = parser.parse("../smt2/quickstart.smt2")
    # We expect no error to occur.
    assert not res

    # Now we retrieve the set of asserted formulas and print them.
    assertions = parser.bitwuzla().get_assertions()
    print('Assertions:')
    print('{')
    for a in assertions:
        print(f' {a}')
    print('}')

    # Now we add an assertion via parsing from string.
    parser.parse('(assert (distinct (select a x) y))', True, False)
    # Now the formula is unsat.
    print('Expect: unsat')
    print(f'Bitwuzla: {parser.bitwuzla().check_sat()}')

    # For illustration purposes, we now parse in some declarations and terms 
    # and sorts from string.

    # Declare bit-vector sort of size 16.
    bv16 = parser.parse_sort('(_ BitVec 16)')
    # Create bit-vector sort of size 16 and show that it corresponds to
    # its string representation '(_ BitVec16)'.
    assert bv16 == tm.mk_bv_sort(16)

    # Declare Boolean constants 'c' and 'd'.
    # Note: Declarations are commands (not terms) in the SMT-LIB language.
    #       Commands must be parsed in via Parser.parse(),
    #       Parser::parse_term() only supports parsing SMT-LIB terms.
    parser.parse("(declare-const c Bool)(declare-const d Bool)", True, False)
    # Declare bit-vector constant 'b'.
    parser.parse('(declare-const b (_ BitVec 16))', True, False)
    # Retrieve term representing 'b'.
    b = parser.parse_term('b')
    # Retrieve term representing 'c'.
    c = parser.parse_term('c')
    # Retrieve term representing 'd'.
    d = parser.parse_term('d')
    # Create xor over 'c' and 'd' and show that it corresponds to term
    # parsed in from its string representation '(xor c d)'.
    assert parser.parse_term('(xor c d)') == tm.mk_term(Kind.XOR, [c, d])
    # Create bit-vector addition over 'b' and bit-vector value
    # '1011111010001010' and show that it corresponds to the term parsed in
    # from its string representation '(bvadd b #b1011111010001010)'.
    assert parser.parse_term('(bvadd b #b1011111010001010)') \
           == tm.mk_term(Kind.BV_ADD,
                      [b, tm.mk_bv_value(bv16, '1011111010001010', 2)])



###
# Bitwuzla: Satisfiability Modulo Theories (SMT) solver.
#
# Copyright (C) 2026 by the authors listed in the AUTHORS file at
# https:#github.com/bitwuzla/bitwuzla/blob/main/AUTHORS
#
# This file is part of Bitwuzla under the MIT license. See COPYING for more
# information at https:#github.com/bitwuzla/bitwuzla/blob/main/COPYING
##

from bitwuzla import *

if __name__ == '__main__':

    tm = TermManager()
    options = Options()
    options.set(Option.PRODUCE_INTERPOLANTS, True)
    options.set(Option.INTERPOLANTS_SIMP, True)
    bitwuzla = Bitwuzla(tm, options)

    bv2 = tm.mk_bv_sort(2)
    bv4 = tm.mk_bv_sort(4)

    x1  = tm.mk_const(bv2, 'x1')
    x2  = tm.mk_const(bv2, 'x2')
    x3  = tm.mk_const(bv2, 'x3')
    a1  = tm.mk_term(Kind.BV_SLT,
                        [tm.mk_bv_zero(bv4),
                         tm.mk_term(Kind.BV_SUB,
                                    [tm.mk_term(Kind.BV_ZERO_EXTEND, [x1], [2]),
                                     tm.mk_bv_one(bv4)])])
    a2  = tm.mk_term(Kind.EQUAL, [x1, x2])
    a3  = tm.mk_term(
            Kind.EQUAL,
            [x3,
             tm.mk_term(Kind.BV_EXTRACT,
                   [tm.mk_term(Kind.BV_NEG,
                               [tm.mk_term(Kind.BV_ZERO_EXTEND, [x2], [2])])],
                   [1, 0])])
    a4 = tm.mk_term(Kind.EQUAL, [x3, tm.mk_bv_zero(bv2)])

    bitwuzla.assert_formula(a1)
    bitwuzla.assert_formula(a2)
    bitwuzla.assert_formula(a3)
    bitwuzla.assert_formula(a4)
    bitwuzla.check_sat()

    # Query an interpolant for A /\ B with A = {a1, a2}.
    interpolant = bitwuzla.get_interpolant([a1, a2])
    # (not (= x2 #b00))
    print(interpolant)

    # Query an interpolation sequence for a sequence of A/B-partitions
    #     {
    #       ({a1},        {a2, a3, a4}),
    #       ({a1, a2},    {a3, a4}),
    #       ({a1, a2, a3, {a4})
    #     }
    # given as a list {{a1}, {a2}, {a3}} of increments of A.
    interpolants = bitwuzla.get_interpolants([[a1], [a2], [a3]])
    # [0]: (= ((_ extract 3 3) (bvadd (concat #b00 x1) #b1111)) #b0)
    # [1]: (not (= x2 #b00))
    # [2]: (not (= x3 #b00))
    print('(')
    for itp in interpolants: print(itp)
    print(')')
